from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from tspace.server.util import AutoIncrementId
from tspace.server.models import DamageType, Countermeasure

if TYPE_CHECKING:
    from tspace.server.models import Galaxy

autoid = AutoIncrementId()

PLASTIC_CHAFF = Countermeasure(
    id=-1,
    name="Plastic Chaff",
    strengths={DamageType.ENERGY},
    weaknesses={DamageType.EXPLOSIVE},
    strength_bonus=2,
    weakness_penalty=2,
)
METAL_SHRAPNEL = Countermeasure(
    id=-1,
    name="Plastic Shrapnel",
    strengths={DamageType.EXPLOSIVE},
    weaknesses={DamageType.KINETIC},
    strength_bonus=2,
    weakness_penalty=2,
)
MANEUVERING_THRUSTERS = Countermeasure(
    id=-1,
    name="Maneuvering Thrusters",
    strengths={DamageType.KINETIC},
    weaknesses={DamageType.ENERGY},
    strength_bonus=2,
    weakness_penalty=2,
)


def create(game: Galaxy, strength_type: DamageType) -> Countermeasure:
    if strength_type == DamageType.ENERGY:
        template = PLASTIC_CHAFF
    elif strength_type == DamageType.KINETIC:
        template = MANEUVERING_THRUSTERS
    elif strength_type == DamageType.EXPLOSIVE:
        template = METAL_SHRAPNEL
    else:
        breakpoint()
        raise ValueError()
    countermeasure = dataclasses.replace(template, id=autoid.incr())
    game.countermeasures[countermeasure.id] = countermeasure
    return countermeasure



from dataclasses import dataclass
from typing import Any, Tuple, List, Callable

from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.widgets import Frame

from tspace.client.ui.dynamic_label import DynamicLabel


@dataclass
class Stat:
    title: str
    callable: Callable[[], Any]
    multiline: bool = False


class StatFrame(Frame):
    def __init__(
        self, title, condition: Callable[[], bool], stats: List[Stat], **kwargs
    ):
        self.condition = condition
        self.stats = stats
        Frame.__init__(self, title=title, body=DynamicLabel(self.get_label), **kwargs)

    def get_label(self):
        render_values = self.condition()
        fragments = []
        max_length = max(len(stat.title) for stat in self.stats if not stat.multiline)
        for stat in self.stats:
            pad = " " * (max_length - len(stat.title))
            fragments.append(("grey", f"{stat.title}: {pad}"))

            value = stat.callable() if render_values else ""
            if isinstance(value, (list, set)):
                fragments.append(("", "\n"))
                value = "\n".join(f" - {part}" for part in value)
                fragments.append(("bold", value))
            else:
                fragments.append(("bold", str(value)))
            fragments.append(("", "\n"))
        return FormattedText(fragments)



from asyncio import Queue
from dataclasses import astuple
from dataclasses import dataclass
from typing import Iterable
from typing import Tuple, Callable, List, Sequence

from prompt_toolkit import ANSI
from prompt_toolkit.formatted_text import to_formatted_text

from tspace.client.instant_cmd import InstantCmd, InvalidSelectionError
from tspace.client.twbuffer import TwBuffer


class Terminal:
    def __init__(self, buffer: TwBuffer):
        self.buffer = buffer
        self.stdin = input

    def backspace(self, num_characters: int = 1):
        self.buffer.backspace(num_characters)

    def write_line(self, *fragments: Tuple[str, str]):
        self.write_lines([*fragments])

    def write_lines(self, *text: Sequence[Tuple[str, str]]):
        self.buffer.insert_after(*text)

    def print(self, text: str, color=None, bg=None, attrs=None):
        style = ""
        if color:
            style += color
        if bg:
            style += f" bg:{bg}"

        if attrs:
            style += " "
            style += " ".join(attrs)

        lines = text.split(r"\n|\r|\r\n")
        self.write_lines(*[[(style, line)] for line in lines])

    def nl(self, times=1):
        for x in range(times):
            self.write_lines([], [])

    def error(self, msg):
        self.nl()
        self.print(msg, "red")
        self.nl()
        self.nl()

    async def read_key(self) -> str:
        queue = Queue()

        def receive_input(txt: str):
            queue.put_nowait(txt)

        self.buffer.input_listeners.append(receive_input)
        return await queue.get()

    # def read_line(self, matcher: Callable[[str],bool]) -> str:
    #     queue = Queue()
    #
    #     buffer = []
    #
    #     def receive_input(txt: str):
    #         if txt == '\r' or txt == '\n':
    #             queue.put_nowait("".join(buffer))
    #         elif matcher(txt):
    #             buffer.append(txt)
    #
    #     self.buffer.input_listeners.append(receive_input)
    #     return queue.get()

    def write_ansi(self, text):
        for line in text.split("\n"):
            self.write_line(*to_formatted_text(ANSI(line)))
            self.nl()


@dataclass
class Fragment:
    style: str
    text: str


@dataclass
class Item:
    value: List[Fragment]


@dataclass
class Row:
    header: Fragment
    items: List[Item]


@dataclass
class Table:
    rows: List[Row]


def print_grid(stream: Terminal, data: Table, separator: Fragment):
    header_len = max(len(row.header.text) for row in data.rows) + 1
    for row in data.rows:
        stream.write_line(
            astuple(row.header),
            ("", " " * (header_len - len(row.header.text))),
            astuple(separator),
        )
        pad = False
        for item in row.items:
            if pad:
                stream.write_line(("", " " * (header_len + 2)))
            else:
                pad = True

            for frag in item.value:
                stream.write_line(astuple(frag))

            stream.nl()


def print_action(stream: Terminal, title):
    stream.nl()
    stream.print(text=title, color="white", bg="blue")
    stream.nl(2)


@dataclass
class Option:
    key: str
    description: str
    function: Callable[[], None]


class SimpleMenuCmd:
    def __init__(self, stream: Terminal, default: str, option_order: Iterable[str]):
        self.stream = stream
        self.default = default
        self.options = {}

        self.instant_prompt = InstantCmd(stream)
        for key in option_order:
            key = key.lower()
            fn = getattr(self, "do_{}".format(key))
            self.options[key] = Option(
                key=key, description=fn.__doc__.strip(), function=fn
            )
            self.instant_prompt.literal(key, key.upper() == default.upper())(fn)

    async def cmdloop(self):
        self.stream.nl()
        for opt in self.options.values():
            self.stream.write_line(
                ("magenta", "<"),
                ("green", opt.key.upper()),
                ("magenta", ">"),
                ("green", opt.description),
            )
            self.stream.nl()
        self.stream.nl()
        while True:
            self.stream.write_line(
                ("magenta", "Enter your choice "),
                ("bold yellow", f"[{self.default.upper()}] "),
            )
            try:
                await self.instant_prompt.cmdloop()
                break
            except InvalidSelectionError:
                self.stream.error("Not a valid option")
            #
            # self.stream.out.flush()
            # val = self.stream.stdin.readline().strip()
            # if val == "":
            #     val = self.default
            #
            # val = val.lower()
            #
            # if val not in self.options:
            #     self.stream.error("Not a valid option")
            # else:
            #     opt = self.options[val]
            #     opt.function()
            #     break


async def menu_prompt(stream: Terminal, default: str, options: Tuple[Tuple[str, str]]):
    for cmd, text in options:
        stream.write_line(
            ("magenta", "<"), ("green", cmd), ("magenta", "> "), ("green", text)
        )
        stream.nl()
    stream.nl()

    while True:
        stream.write_line(
            ("magenta", "Enter your choice "), ("bold yellow", f"[{default}] ")
        )
        val = await stream.read_key()
        if val == "":
            val = default

        if val not in [o[0] for o in options]:
            stream.error("Not a valid option")
        else:
            return val


async def amount_prompt(
    stream: Terminal,
    prompt: Sequence[Tuple[str, str]],
    default: int,
    min: int,
    max: int,
):
    _write_prompt_and_default(default, prompt, stream)
    prompt = InstantCmd(stream)

    async def handle_input(txt):
        value = int(txt)
        if value < min or value > max:
            raise ValueError("Number out of range")
        else:
            return value

    prompt.regex("[0-9]+", max_length=len(str(max)))(handle_input)
    prompt.literal("y", default=True)(lambda _: default)
    return await prompt.cmdloop()


def _write_prompt_and_default(default, prompt, stream):
    stream.write_line(*prompt)
    stream.write_line(
        ("magenta", "["), ("yellow", str(default)), ("magenta", "]?"), ("", " ")
    )


async def yesno_prompt(
    stream: Terminal, prompt: Sequence[Tuple[str, str]], default: bool, **kwargs
):
    _write_prompt_and_default("Y" if default else "N", prompt, stream)

    val = await stream.read_key()
    if not val.strip():
        val = default
    return val is True or "y" == val.lower()



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from tspace.server.models import Weapon, DamageType
from tspace.server.util import AutoIncrementId

if TYPE_CHECKING:
    from tspace.server.models import Galaxy

autoid = AutoIncrementId()

MINING_LASER = Weapon(
    id=-1,
    name="Mining Laser",
    accuracy_bonus=1,
    damage_type=DamageType.ENERGY,
    damage_max=3,
    damage_min=1,
)
CONCUSSION_MISSILE = Weapon(
    id=-1,
    name="Concussion Missile",
    accuracy_bonus=1,
    damage_type=DamageType.EXPLOSIVE,
    damage_max=3,
    damage_min=1,
)
RAILGUN = Weapon(
    id=-1,
    name="Railgun",
    accuracy_bonus=1,
    damage_type=DamageType.KINETIC,
    damage_max=3,
    damage_min=1,
)


def create(game: Galaxy, type: DamageType) -> Weapon:
    if type == DamageType.ENERGY:
        weapon_template = MINING_LASER
    elif type == DamageType.KINETIC:
        weapon_template = RAILGUN
    elif type == DamageType.EXPLOSIVE:
        weapon_template = CONCUSSION_MISSILE
    else:
        breakpoint()
        raise ValueError()
    weapon = dataclasses.replace(weapon_template, id=autoid.incr())
    game.weapons[weapon.id] = weapon
    return weapon



from __future__ import annotations
from tspace.server.constants import PORT_NAMES, PLANET_SUFFIXES
from typing import TYPE_CHECKING, Optional

from tspace.server.util import AutoIncrementId
from tspace.server.models import Planet

if TYPE_CHECKING:
    from tspace.server.models import Galaxy

autoid = AutoIncrementId()


def create(game: Galaxy, owner_id: Optional[int]) -> Planet:
    base_name = game.rnd.choice(PORT_NAMES)
    suffix = game.rnd.choice(PLANET_SUFFIXES)
    name = f"{base_name} {suffix}".strip()
    type = game.rnd.choice(["V", "M"])
    planet = Planet(game, autoid.incr(), type, name, owner_id)
    game.planets[planet.id] = planet
    return planet



from __future__ import annotations

from typing import Tuple

from colorclass import Color
from tabulate import DataRow
from tabulate import Line
from tabulate import TableFormat
from tabulate import tabulate

from tspace.client.game import Game
from tspace.client.models import PlayerClient
from tspace.client.models import Port
from tspace.client.models import PortClient
from tspace.client.models import Ship
from tspace.client.prompts import PromptTransition
from tspace.client.prompts import PromptType
from tspace.client.stream import SimpleMenuCmd
from tspace.client.stream import Terminal
from tspace.client.stream import amount_prompt
from tspace.client.stream import print_action
from tspace.client.stream import yesno_prompt


class Actions:
    async def buy_from_port(
        self, port_id: int, commodity: str, amount: int
    ) -> Tuple[PlayerClient, PortClient]:
        pass

    async def sell_to_port(
        self, port_id: int, commodity: str, amount: int
    ) -> Tuple[PlayerClient, PortClient]:
        pass

    async def exit_port(self, port_id: int):
        pass


class Prompt(SimpleMenuCmd):
    def __init__(self, game: Game, actions: Actions, term: Terminal):
        super().__init__(term, "T", ("T", "Q"))
        self.out = term
        self.game = game
        self.player = game.player
        self.actions = actions
        self.out.nl()

    async def do_t(self, _):
        """
        Trade at this Port
        """
        await self._print_table(self.player.ship, self.player.port)
        raise PromptTransition(PromptType.SECTOR)

    async def do_q(self, _):
        """
        Quit, nevermind
        """
        self.out.nl(2)
        await self.actions.exit_port(port_id=self.player.port_id)
        raise PromptTransition(PromptType.NONE)

    async def _print_table(self, ship: Ship, port: Port):
        print_action(self.out, "Port")
        self.out.write_line(
            ("yellow", "Commerce report for "),
            ("cyan", port.name),
            ("yellow", ": 12:50:33 PM Sat May 06, 2028"),
        )
        self.out.nl()
        self.out.write_line(("magenta", "-=-=-        Docking Log        -=-=-"))
        self.out.nl(2)
        self.out.write_line(("green", "No current ship docking log on file."))
        self.out.nl(2)
        rows = []
        for c in port.commodities.values():
            rows.append(
                [
                    Color.cyan("{name}").format(name=c.type.value),
                    Color.green("Buying" if c.buying else "Selling"),
                    Color.cyan("{}").format(c.amount),
                    Color("{green}{}{/green}{red}%{/red}").format(
                        int(c.amount / c.capacity * 100)
                    ),
                    Color.cyan(str(ship.holds[c.type])),
                ]
            )

        self.out.write_ansi(
            tabulate(
                tabular_data=rows,
                stralign="center",
                numalign="center",
                headers=(
                    Color.green(t)
                    for t in ["Items", "Status", "Trading", "% of max", "OnBoard"]
                ),
                tablefmt=TableFormat(
                    lineabove=Line("", "-", "  ", ""),
                    linebelowheader=Line("", Color.magenta("-"), "  ", ""),
                    linebetweenrows=None,
                    linebelow=Line("", "-", "  ", ""),
                    headerrow=DataRow("", "  ", ""),
                    datarow=DataRow("", "  ", ""),
                    padding=0,
                    with_header_hide=["lineabove", "linebelow"],
                ),
            )
        )
        self.out.nl(2)

        self.print_trader_status()

        if self.player.ship.holds_free < self.player.ship.holds_capacity:
            for c in port.commodities.values():
                in_holds = self.player.ship.holds.get(c.type, 0)
                if c.buying and c.amount and in_holds:
                    amount = min(c.amount, in_holds)
                    self.out.write_line(
                        ("magenta", "We are buying up to "),
                        ("yellow", str(c.amount)),
                        ("magenta", ". You have "),
                        ("yellow", str(self.player.ship.holds[c.type])),
                        ("magenta", "in your holds. "),
                    )
                    self.out.nl()
                    value = await amount_prompt(
                        stream=self.out,
                        prompt=(
                            ("magenta", "How many holds of "),
                            ("cyan", c.type.value),
                            ("magenta", " do you want to sell"),
                        ),
                        default=amount,
                        min=0,
                        max=amount,
                    )
                    if value:
                        self.out.nl()
                        self.out.write_line(
                            ("cyan", "Agreed, "),
                            ("yellow", str(value)),
                            ("cyan", " units."),
                        )
                        self.out.nl(2)
                        if await yesno_prompt(
                            self.out,
                            prompt=(
                                ("green", "We'll buy them for "),
                                ("yellow", str(int(c.price * value))),
                                ("green", " credits."),
                            ),
                            default=True,
                        ):
                            player_client, port_client = await self.actions.sell_to_port(
                                port_id=port.id, commodity=c.type, amount=value
                            )
                            self.game.update_player(player_client)
                            self.game.update_port(port_client)

                            self.out.nl()
                            self.out.print(
                                "You drive a hard bargain, but we'll take them.",
                                color="magenta",
                            )
                            self.out.nl(2)

        if self.player.ship.holds_free and sum(
            [c.amount for c in port.commodities.values() if not c.buying]
        ):
            for c in port.commodities.values():
                amount = min(self.player.ship.holds_free, c.amount)
                if not c.buying and amount:
                    existing = self.player.ship.holds.get(c.type, 0)

                    self.out.write_line(
                        ("magenta", "We are selling up to "),
                        ("yellow", str(c.amount)),
                        ("magenta", ". You have "),
                        ("yellow", str(existing)),
                        ("magenta", " in your holds."),
                    )
                    self.out.nl()
                    value = await amount_prompt(
                        stream=self.out,
                        prompt=[
                            ("magenta", "How many holds of "),
                            ("cyan", c.type.value),
                            ("magenta", " do you want to buy"),
                        ],
                        default=amount,
                        min=0,
                        max=amount,
                    )
                    if value:
                        self.out.nl()
                        self.out.write_line(
                            ("cyan", "Agreed, "),
                            ("yellow", str(value)),
                            ("cyan", " units."),
                        )
                        self.out.nl(2)
                        if await yesno_prompt(
                            self.out,
                            prompt=[
                                ("green", "We'll sell them for "),
                                ("yellow", str(int(c.price * value))),
                                ("green", " credits."),
                            ],
                            default=True,
                        ):
                            player_client, port_client = await self.actions.buy_from_port(
                                port_id=port.id, commodity=c.type, amount=value
                            )
                            self.game.update_player(player_client)
                            self.game.update_port(port_client)

                            self.out.nl()
                            self.out.print(
                                "You are a shrewd trader, they're all yours.",
                                color="green",
                            )
                            self.out.nl(2)
                        else:
                            self.out.nl(2)
                            self.out.print("Fine, leave then.", color="red")
                            self.out.nl(2)

                    self.out.nl()
                    self.print_trader_status()

                    # self.out.print(" Items     Status  Trading % of max OnBoard", 'green')
                    # self.out.nl()
                    # self.out.print(" -----     ------  ------- -------- -------", 'magenta')
                    # for c in port.commodities:
                    #     self.out.print(c.type.replace('_', ' ').title(), 'cyan', attrs=['bold'])
                    #     self.out.write(" " * (12 - len(c.type)))
                    #     self.out.print("Buying" if c.buying else "Selling")
                    #     self.out.write(" " * (26 - ))

    def print_trader_status(self):
        self.out.write_line(
            ("green", "You have "),
            ("yellow", str(self.player.credits)),
            ("green", " credits and "),
            ("yellow", str(self.player.ship.holds_free)),
            ("green", " empty cargo holds"),
        )
        self.out.nl(2)


class Events:
    def __init__(self, prompt: Prompt):
        self.prompt = prompt

    # noinspection PyMethodMayBeStatic
    def on_invalid_action(self, error: str):
        self.prompt.out.error(error)

    def on_port_buy(self, id: int, player: PlayerClient):
        self.prompt.player.update(player)
        # todo: when async, make this sync on event id

    def on_port_sell(self, id: int, player: PlayerClient):
        self.prompt.player.update(player)
        # todo: when async, make this sync on event id



PORT_NAMES = [
    "Aegis",
    "Aeon",
    "Aeris",
    "Babylon",
    "Aeternitas",
    "Aether",
    "Alliance",
    "Alpha",
    "Amazone",
    "Ancestor",
    "Anemone",
    "Angel",
    "Anomaly",
    "Apollo",
    "Arcadia",
    "Arcadis",
    "Arch",
    "Architect",
    "Ark",
    "Artemis",
    "Asphodel",
    "Asteria",
    "Astraeus",
    "Athena",
    "Atlas",
    "Atmos",
    "Aura",
    "Aurora",
    "Awe",
    "Azura",
    "Azure",
    "Baldur",
    "Beacon",
    "Blue Moon",
    "Borealis",
    "Burrow",
    "Caelestis",
    "Canaan",
    "Century",
    "Chrono",
    "Chronos",
    "Crescent",
    "Curator",
    "Curiosity",
    "Data",
    "Dawn",
    "Daydream",
    "Demeter",
    "Dogma",
    "Dream",
    "Dune",
    "Ecstacys",
    "Eir",
    "Elyse",
    "Elysium",
    "Empyrea",
    "Ender",
    "Enigma",
    "Eos",
    "Epiphany",
    "Epitome",
    "Erebus",
    "Escort",
    "Eternis",
    "Eternity",
    "Exposure",
    "Fable",
    "Father",
    "Fauna",
    "Felicity",
    "Flora",
    "Fortuna",
    "Frontier",
    "Gaia",
    "Galaxy",
    "Genesis",
    "Genius",
    "Glory",
    "Guardian",
    "Halo",
    "Heirloom",
    "Helios",
    "Hemera",
    "Hera",
    "Heritage",
    "Hermes",
    "Horus",
    "Hymn",
    "Hyperion",
    "Hypnos",
    "Ignis",
    "Illume",
    "Inception",
    "Infinity",
    "Isis",
    "Janus",
    "Juno",
    "Legacy",
    "Liberty",
    "Lore",
    "Lucent",
    "Lumina",
    "Luminous",
    "Luna",
    "Lunis",
    "Magni",
    "Mammoth",
    "Mani",
    "Marvel",
    "Memento",
    "Minerva",
    "Miracle",
    "Mother",
    "Muse",
    "Mystery",
    "Mythos",
    "Nebula",
    "Nemesis",
    "Nemo",
    "Neo",
    "Nero",
    "Nimbus",
    "Nott",
    "Nova",
    "Novis",
    "Nox",
    "Nyx",
    "Odyssey",
    "Olympus",
    "Omega",
    "Oracle",
    "Orbital",
    "Origin",
    "Orphan",
    "Osiris",
    "Outlander",
    "Parable",
    "Paradox",
    "Paragon",
    "Pedigree",
    "Phantasm",
    "Phantom",
    "Phenomenon",
    "Phoenix",
    "Pilgrim",
    "Pioneer",
    "Prism",
    "Prodigy",
    "Prometheus",
    "Prophecy",
    "Proto",
    "Radiance",
    "Rebus",
    "Relic",
    "Revelation",
    "Reverie",
    "Rogue",
    "Rune",
    "Saga",
    "Sancus",
    "Scout",
    "Selene",
    "Serenity",
    "Settler",
    "Shangris",
    "Shepherd",
    "Shu",
    "Sol",
    "Solas",
    "Spectacle",
    "Specter",
    "Spectrum",
    "Spire",
    "Symbolica",
    "Tartarus",
    "Terminus",
    "Terra",
    "Terran",
    "Terraria",
    "Themis",
    "Tiberius",
    "Titan",
    "Titanus",
    "Torus",
    "Tranquility",
    "Trivia",
    "Utopis",
    "Valhalla",
    "Vanguard",
    "Vanquish",
    "Vesta",
    "Vestige",
    "Victoria",
    "Virtue",
    "Visage",
    "Voyage",
    "Vulcan",
    "Warden",
    "Yggdrasil",
    "Zeus",
    "Zion",
]
PORT_SUFFIXES = ["", "Station", "Base", "Terminal", "Outpost"] + [
    str(x) for x in range(0, 9)
]
PLANET_SUFFIXES = [
    "II",
    "III",
    "IV",
    "V",
    "V1",
    "VII",
    "VIII",
    "IX",
    "Alpha",
    "Beta",
    "Omega",
    "Gamma",
    "Zeta",
]



from enum import Enum


class PromptType(Enum):
    SECTOR = 1
    PORT = 2
    QUIT = 3
    NONE = 4


class PromptTransition(Exception):
    def __init__(self, next_prompt: PromptType):
        self.next = next_prompt



from typing import Any, Optional


class PortConfig:
    def __init__(self, density: int = 40):
        self.density = density


class PlayerConfig:
    def __init__(self, initial_sector_id: int = 1, initial_credits: int = 2000):
        self.initial_credits = initial_credits
        self.initial_sector_id = initial_sector_id


class GameConfig:
    def __init__(
        self,
        id: int,
        name: str,
        diameter: int = 50,
        seed: Optional[Any] = None,
        port: PortConfig = PortConfig(),
        player: PlayerConfig = PlayerConfig(),
        debug_network: bool = False,
        warp_density: int = 3.5,
    ):
        self.player = player
        self.warp_density = warp_density
        self.debug_network = debug_network
        self.seed = seed
        self.port = port
        self.diameter = diameter
        self.name = name
        self.id = id



from asyncio import Future
from functools import partial

from prompt_toolkit import Application
from prompt_toolkit.layout import Layout
from prompt_toolkit.widgets import Button
from prompt_toolkit.widgets import Label

from tspace.client.ui.menu import MenuDialog
from tspace.client.ui.starfield import Starfield


class TitleScene:
    def __init__(self, app: Application):
        self.app = app

        dialog = MenuDialog(
            title="Terminal Space",
            body=Label(
                text="   A text-based\n space trading game",
                dont_extend_width=False,
                width=20,
                dont_extend_height=True,
            ),
            buttons=[
                Button(text="New Game", handler=partial(self.do_exit, "start")),
                Button(text="Join Game", handler=partial(self.do_exit, "join")),
                Button(text="Quit", handler=partial(self.do_exit, "quit")),
            ],
            background=Starfield(),
            on_dismiss=partial(self.do_exit, "quit"),
        )

        self.layout = Layout(dialog)
        self.future = Future()

    def do_exit(self, result, *_):
        self.future.set_result(result)

    async def start(self):
        result = await self.future
        self.future = Future()
        return result

    def end(self):
        pass



from typing import Callable
from typing import Tuple, List, Sequence

from prompt_toolkit.layout.screen import Point


class TwBuffer:
    def __init__(self):
        self.buffer: List[List[Tuple[str, str]]] = [[]]
        self.cursor = Point(x=0, y=0)

        self.input_listeners: List[Callable[[str], None]] = []
        self.change_listeners: List[Callable[[], None]] = []

    def insert_after(self, *text: Sequence[Tuple[str, str]]):
        if not isinstance(text, (Tuple, List)):
            raise ValueError()

        if any(isinstance(x, Tuple) for x in text):
            raise ValueError()

        str_buffer = []
        for line in text:
            line_buffer = []
            for frag in line:
                if not isinstance(frag, Tuple):

                    raise ValueError(f"Invalid fragment: {frag}")
                if not isinstance(frag[1], str):
                    line_buffer.append((frag[0], str(frag[1])))
                else:
                    line_buffer.append(frag)
            str_buffer.append(line_buffer)
        text = str_buffer

        if text and text[0]:
            self.buffer[self.cursor.y] += text[0]

        if len(text) > 1:
            self.buffer += text[1:]

        self._set_cursor_to_buffer_end()

        for listener in self.change_listeners:
            listener()

    def _set_cursor_to_buffer_end(self):
        new_y = len(self.buffer) - 1
        new_x = self.get_line_length(new_y) - 1
        self.cursor = Point(x=new_x, y=new_y)

    def on_change(self, listener: Callable[[], None]):
        self.change_listeners.append(listener)

    def get_line(self, index: int) -> List[Tuple[str, str]]:
        result = [] if index >= len(self.buffer) else self.buffer[index]
        return result + [("", " ")]

    def get_line_length(self, index: int) -> int:
        return sum(len(x[1]) for x in self.get_line(index))

    @property
    def line_count(self) -> int:
        return len(self.buffer)

    @property
    def cursor_position(self) -> Point:
        return self.cursor

    def on_key_press(self, txt: str):
        for listener in self.input_listeners:
            listener(txt)

    def backspace(self, num_characters: int):
        line = self.buffer[-1]
        if num_characters >= self.get_line_length(-1):
            self.buffer[-1] = []
        else:
            while num_characters:
                last = line[-1]
                chars = last[1]
                if len(chars) >= num_characters:
                    num_characters = 0
                    chars = chars[0:-num_characters]
                else:
                    num_characters -= len(chars)
                    chars = ""
                line[-1] = (last[0], chars)

        self._set_cursor_to_buffer_end()



from __future__ import annotations
from typing import Dict
from typing import List

import networkx as nx

from tspace.client.models import GameConfig, WeaponClient
from tspace.client.models import Planet
from tspace.client.models import PlanetClient
from tspace.client.models import Player
from tspace.client.models import PlayerClient
from tspace.client.models import Port
from tspace.client.models import PortClient
from tspace.client.models import Sector
from tspace.client.models import SectorClient
from tspace.client.models import Ship
from tspace.client.models import ShipClient
from tspace.client.models import Trader
from tspace.client.models import TraderClient
from tspace.client.models import TraderShip
from tspace.client.models import TraderShipClient


class Game:
    def __init__(self, config: GameConfig):
        self.config = config
        self.sectors: List[Sector] = [None] * (config.sectors_count + 1)
        self.trader_ships: Dict[int, TraderShip] = {}
        self.ships: Dict[int, Ship] = {}
        self.ports: Dict[int, Port] = {}
        self.traders: Dict[int, Trader] = {}
        self.planets: Dict[int, Planet] = {}
        self.player: Player = None

    # noinspection PyUnresolvedReferences
    def update_player(self, client: PlayerClient) -> Player:
        if self.player:
            self.player.update(client)
        else:
            self.player = Player(self, client)

        if client.ship:
            self.update_ship(client.ship)

        return self.player

    def update_ship(self, client: ShipClient) -> Ship:
        if client.id in self.ships:
            self.ships[client.id].update(client)
        else:
            self.ships[client.id] = Ship(self, client)

        if client.sector:
            self.update_sector(client.sector)
            self.ships[client.id].sector_id = client.sector.id
            self.player.visited.add(client.sector.id)

        return self.ships[client.id]

    def update_sector(self, client: SectorClient) -> Sector:
        sector = self.sectors[client.id] if client.id in self.sectors else None
        if sector:
            sector.update(client)
        else:
            self.sectors[client.id] = Sector(self, client)

        if client.ports:
            for port in client.ports:
                self.update_port(port)

        if client.ships:
            for ship in client.ships:
                self.update_trader_ship(ship)

        if client.planets:
            for planet in client.planets:
                self.update_planet(planet)

        for warp_id in (x for x in client.warps if not self.sectors[x]):
            self.sectors[warp_id] = Sector(
                self, SectorClient(id=warp_id, warps=[], ports=[], ships=[], planets=[])
            )

        return self.sectors[client.id]

    def update_trader_ship(self, client: TraderShipClient) -> TraderShip:
        if client.id in self.trader_ships:
            self.trader_ships[client.id].update(client)
        else:
            self.trader_ships[client.id] = TraderShip(self, client)

        if client.trader:
            self.update_trader(client.trader)

        return self.trader_ships[client.id]

    def update_port(self, client: PortClient) -> Port:
        if client.id in self.ports:
            self.ports[client.id].update(client)
        else:
            self.ports[client.id] = Port(self, client)

        return self.ports[client.id]

    def update_trader(self, client: TraderClient) -> Trader:
        if client.id in self.traders:
            self.traders[client.id].update(client)
        else:
            self.traders[client.id] = Trader(self, client)

        return self.traders[client.id]

    def update_planet(self, client: PlanetClient) -> Planet:
        if client.id in self.planets:
            self.planets[client.id].update(client)
        else:
            self.planets[client.id] = Planet(self, client)

        return self.planets[client.id]

    def plot_course(self, from_id: int, to_id: int) -> List[Sector]:
        g = self._gen_graph()
        steps: List[int] = nx.shortest_path(g, from_id, to_id)
        return [self.sectors[x] for x in steps]

    def _gen_graph(self) -> nx.Graph:
        g = nx.Graph(directed=True)

        for sector in (sector for sector in self.sectors if sector):
            g.add_node(sector.id)
            for warp in sector.warps:
                g.add_edge(sector.id, warp)

        return g



from __future__ import annotations

import random
from typing import TYPE_CHECKING

from tspace.server.constants import PORT_NAMES, PORT_SUFFIXES
from tspace.server.util import AutoIncrementId
from tspace.server.models import Port, PortClass, CommodityType, TradingCommodity

if TYPE_CHECKING:
    from tspace.server.models import Galaxy


def random_port_type(rnd: random.Random) -> PortClass:
    type_chance = rnd.randint(0, 99)
    if type_chance < 20:
        return PortClass.by_id(1)
    elif type_chance < 40:
        return PortClass.by_id(2)
    elif type_chance < 60:
        return PortClass.by_id(3)
    elif type_chance < 70:
        return PortClass.by_id(4)
    elif type_chance < 80:
        return PortClass.by_id(5)
    elif type_chance < 90:
        return PortClass.by_id(6)
    elif type_chance < 95:
        return PortClass.by_id(7)
    else:
        return PortClass.by_id(8)


autoid = AutoIncrementId()


def create(game: Galaxy, sector_id: int) -> Port:
    commodities = []

    ptype = random_port_type(game.rnd)

    for ctype in CommodityType:
        buying = ptype.buying[ctype]
        amount = game.rnd.randint(200, 2000)
        commodities.append(TradingCommodity(ctype, amount, buying))

    name = game.rnd.choice(PORT_NAMES)
    suffix = game.rnd.choice(PORT_SUFFIXES)
    if suffix:
        name = " ".join([name, suffix])
        name = name.strip()
    port = Port(autoid.incr(), sector_id, name, commodities)
    game.ports[port.id] = port
    return port



import inspect
import json
import typing
from enum import Enum
from functools import partial


class TypedEncoder(json.JSONEncoder):
    def __init__(self, *args, exclude_none=False, set_as_list=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.exclude_none = exclude_none
        self.set_as_list = set_as_list

    def default(self, o):

        if isinstance(o, Enum):
            return o.name

        result = o.__dict__
        if self.set_as_list:
            result = {
                k: v if not isinstance(v, set) else list(v)
                for k, v in o.__dict__.items()
            }

        if self.exclude_none:
            result = {k: v for k, v in result.items() if v is not None}

        return result


def decodes(data, cls, **kwargs):
    result = json.loads(data, **kwargs)
    return decode(result, cls)


def decode(data: dict, cls, context=None, **kwargs):
    return _recursive_decode(data, cls, context)


def encodes(obj, exclude_none=False, set_as_list=False, func_kwargs=None, **kwargs):
    return json.dumps(
        obj,
        cls=partial(
            TypedEncoder, set_as_list=set_as_list, exclude_none=exclude_none, **kwargs
        ),
    )


def _recursive_decode(obj, cls, context):

    origin = cls.__origin__ if hasattr(cls, "__origin__") else None
    is_class = False if origin else inspect.isclass(cls)
    if obj is None:
        return obj
    if cls in (str, int, float, bool):
        return obj
    elif is_class and issubclass(cls, Enum):
        return cls[obj]
    elif origin:
        ptypes = cls.__args__
        if str(origin) == "typing.Union":
            if ptypes:
                for ptype in ptypes:
                    try:
                        return _recursive_decode(obj, ptype, context)
                    except AttributeError:
                        pass
        if issubclass(origin, set):
            if ptypes:
                return {_recursive_decode(x, ptypes[0], context) for x in obj}
        elif issubclass(origin, list):
            if ptypes:
                return [_recursive_decode(x, ptypes[0], context) for x in obj]
        elif issubclass(origin, dict):
            if ptypes:
                return {
                    k: _recursive_decode(v, ptypes[1], context) for k, v in obj.items()
                }
        elif issubclass(origin, tuple):
            if ptypes:
                return (
                    _recursive_decode(obj[0], ptypes[0], context),
                    _recursive_decode(obj[1], ptypes[1], context),
                )

    elif is_class:
        try:
            globalns = {"Dict": typing.Dict, "List": typing.List, "Tuple": typing.Tuple}
            hints = typing.get_type_hints(
                cls.__init__, localns=context, globalns=globalns
            )
        except AttributeError:
            raise Exception("Unable to decode class {}".format(cls))
    else:
        hints = typing.get_type_hints(cls, context)

    kwargs = {}
    for prop, ptype in hints.items():
        val = obj.get(prop)
        kwargs[prop] = _recursive_decode(val, ptype, context)

    if not is_class:
        result = kwargs
    else:
        if "return" in kwargs:
            del kwargs["return"]
        result = cls(**kwargs)
    return result



from typing import List

from tspace.client.game import Game
from tspace.client.instant_cmd import InstantCmd
from tspace.client.instant_cmd import InvalidSelectionError
from tspace.client.models import Sector
from tspace.client.models import SectorClient
from tspace.client.models import TraderShipClient
from tspace.client.prompts import PromptTransition
from tspace.client.prompts import PromptType
from tspace.client.stream import Fragment
from tspace.client.stream import Item
from tspace.client.stream import Row
from tspace.client.stream import Table
from tspace.client.stream import Terminal
from tspace.client.stream import print_action
from tspace.client.stream import print_grid


class Actions:
    async def move_trader(self, sector_id: int) -> SectorClient:
        pass

    async def enter_port(self, port_id: int):
        pass


# noinspection PyMethodMayBeStatic,PyIncorrectDocstring
class Prompt:
    def __init__(self, game: Game, actions: Actions, term: Terminal):
        self.out = term
        self.player = game.player
        self.actions = actions
        self.game = game

        self.instant_cmd = InstantCmd(term)
        self.instant_cmd.literal("d", default=True)(self.do_d)
        self.instant_cmd.literal(
            "p", validator=lambda _: len(self.player.sector.ports) > 0
        )(self.do_p)
        self.instant_cmd.literal("q")(self.do_q)
        self.instant_cmd.regex(
            "[0-9]+", max_length=len(str(game.config.sectors_count))
        )(self.do_move)

    async def on_ship_enter_sector(self, sector: SectorClient, ship: TraderShipClient):
        self.game.update_sector(sector)
        if self.player.ship.sector.id == sector.id:
            self.print_ship_enter_sector(ship)

    async def on_ship_exit_sector(self, sector: SectorClient, ship: TraderShipClient):
        self.game.update_sector(sector)
        if self.player.ship.sector.id == sector.id:
            self.print_ship_exit_sector(ship)

    # noinspection PyUnusedLocal
    async def do_d(self, line):
        """
        Re-display the current sector
        """
        print_action(self.out, "Redisplay")
        self.print_sector()

    async def do_p(self, line):
        self.out.nl(2)
        self.out.print("Docking...", color="red", attrs=["blink"])
        await self.actions.enter_port(port_id=self.player.sector.ports[0].id)
        raise PromptTransition(PromptType.NONE)

    def do_q(self, line):
        raise PromptTransition(PromptType.QUIT)

    async def cmdloop(self):
        while True:
            try:
                self.out.write_line(*self.prompt)
                return await self.instant_cmd.cmdloop()
            except InvalidSelectionError:
                self.out.nl()

    @property
    def prompt(self):
        s = self.player.ship.sector
        return (
            ("magenta", "Command ["),
            ("yellow", "TL"),
            ("magenta", "="),
            ("yellow", "00:00:00"),
            ("magenta", "]"),
            ("yellow", ":"),
            ("magenta", "["),
            ("cyan", str(s.id)),
            ("magenta", "] ("),
            ("yellow", "?=Help"),
            ("magenta", ")? : "),
        )

    async def do_move(self, target_sector):
        """
        Move to an adjacent sector
        """
        try:
            target_id = int(target_sector)
        except ValueError:
            self.out.error("Invalid sector number")
            return

        print_action(self.out, "Move")
        if target_id in self.player.sector.warps:
            self.out.write_line(
                ("magenta", "Warping to Sector "), ("yellow", str(target_id))
            )
            self.out.nl(2)

            sector_client = await self.actions.move_trader(sector_id=target_id)
            s = self.game.update_sector(sector_client)

            self.game.player.ship.sector_id = s.id
            self.game.player.visited.add(s.id)
            self.print_sector(s)

        elif self.game.sectors[target_id]:
            self.out.write_line(
                ("magenta", "Auto-warping to Sector "), ("yellow", str(target_id))
            )
            self.out.nl(2)
            warps = self.game.plot_course(self.player.sector.id, target_id)
            for warp in warps[1:]:
                if not warp:
                    breakpoint()

                print_action(self.out, f"Warping to sector {warp.id}")
                sector_client = await self.actions.move_trader(sector_id=warp.id)
                s = self.game.update_sector(sector_client)

                self.game.player.ship.sector_id = s.id
                self.game.player.visited.add(s.id)
                self.print_sector(s)

        else:
            self.out.error("Unknown target sector")

    def print_sector(self, sector: Sector = None):

        if not sector:
            sector = self.player.sector

        data = Table(rows=[])

        data.rows.append(
            Row(
                header=Fragment("green bold", "Sector "),
                items=[
                    Item(
                        [
                            Fragment("cyan", str(sector.id)),
                            Fragment("green", " in "),
                            Fragment("blue", "uncharted space"),
                        ]
                    )
                ],
            )
        )

        if sector.planets:
            items = []
            for planet in sector.planets:
                items.append(
                    Item(
                        [
                            Fragment("fg:yellow bold", f"({planet.planet_type}) "),
                            Fragment("white", planet.name),
                        ]
                    )
                )
            data.rows.append(Row(header=Fragment("magenta", "Planets"), items=items))

        if sector.ports:
            items = []
            for port in sector.ports:
                item = Item(
                    [
                        Fragment("cyan", port.name),
                        Fragment("yellow", ", "),
                        Fragment("magenta", "Class "),
                        Fragment("cyan", str(port.class_number)),
                        Fragment("magenta", " ("),
                    ]
                )
                item.value += port.class_name_colored
                item.value.append(Fragment("magenta", ")"))
                items.append(item)
            data.rows.append(Row(header=Fragment("green", "Port"), items=items))
        #
        # other_ships = [s for s in sector.ships if s.trader.id != self.player.ship.id]
        # if other_ships:
        #     lines = []
        #     for ship in other_ships:
        #         lines.append(Color("{red}{trader}{/red}{yellow},{/yellow}").format(trader=ship.trader.name))
        #         lines.append(Color("{green}in{/green} {cyan}{ship}{/cyan} {green}({ship_type}){/green}").format(
        #                 ship=ship.name,
        #                 ship_type="unknown"
        #         ))
        #     data.append((Color.yellow("Ships"), lines))

        print_grid(self.out, data, separator=Fragment("yellow", ": "))

        warps: List[Fragment] = []
        for w in sector.warps:
            if w in self.player.visited:
                warps.append(Fragment("cyan", str(w)))
            else:
                warps += [
                    Fragment("magenta", "("),
                    Fragment("red", str(w)),
                    Fragment("magenta", ")"),
                ]
            warps.append(Fragment("green", " - "))

        data = Table(
            rows=[
                Row(
                    header=Fragment("green bold", "Warps to Sector(s)"),
                    items=[Item(warps[:-1])],
                )
            ]
        )
        print_grid(self.out, data, separator=Fragment("yellow", ": "))
        self.out.nl()

    def print_ship_enter_sector(self, ship):
        self.out.nl()
        self.out.write_line(
            ("cyan bold", ship.trader.name), ("green", " warps into th esector.")
        )
        self.out.nl()

    def print_ship_exit_sector(self, ship):
        self.out.nl()
        self.out.write_line(
            ("cyan bold", ship.trader.name), ("green", " warps out of the sector.")
        )
        self.out.nl()



import asyncio
import inspect
from asyncio import CancelledError
from asyncio import Future
from asyncio import Task
from typing import Awaitable
from typing import Callable
from typing import Optional

from tspace.client import models, sector_prompt, port_prompt
from tspace.client.game import Game
from tspace.client.instant_cmd import InstantCmd
from tspace.client.models import GameConfig
from tspace.client.models import GameConfigClient
from tspace.client.models import PlayerClient
from tspace.client.models import PortClient
from tspace.client.prompts import PromptTransition
from tspace.client.prompts import PromptType
from tspace.client.stream import Terminal
from tspace.client.util import EventBus


class Session:
    def __init__(self, term: Terminal):
        self.term = term

        self.game: Optional[Game] = None
        self.action_sink = None

        self.bus: Optional[EventBus] = None
        self.prompt = self.start_no_prompt()
        self.prompt_task: Optional[Task] = None

    async def start(self, action_sink: Callable[[str], Awaitable[None]]):
        context = {
            k: v
            for k, v in inspect.getmembers(models, inspect.isclass)
            if k.endswith("Client")
        }
        self.bus = EventBus(self, context=context, sender=action_sink)

        waiting_prompt = None
        while True:
            try:
                if self.prompt is None:
                    break
                waiting_prompt = self.prompt
                self.prompt_task = asyncio.create_task(self.prompt.cmdloop())
                await self.prompt_task
            except CancelledError:
                pass
            except PromptTransition as e:
                self.bus.remove_event_listener(self.prompt)
                if waiting_prompt and waiting_prompt == self.prompt:
                    if e.next == PromptType.SECTOR:
                        self.prompt = self.start_sector_prompt()
                    elif e.next == PromptType.PORT:
                        self.prompt = self.start_port_prompt()
                    elif e.next == PromptType.QUIT:
                        break
                    elif e.next == PromptType.NONE:
                        self.prompt = self.start_no_prompt()

    async def on_game_enter(self, player: PlayerClient, config: GameConfigClient):
        # print("on game enter!!!!!!!!!!!!!")
        self.game = Game(GameConfig(config))
        self.game.update_player(player)

        prompt = self.start_sector_prompt()
        # prompt.print_sector(self.game.player.ship.sector)
        self.prompt = prompt
        self.prompt_task.cancel()

    async def on_port_enter(self, port: PortClient, player: PlayerClient):
        p = self.game.update_port(port)
        self.game.update_player(player)
        self.game.player.port_id = p.id

        self.prompt = self.start_port_prompt()
        self.prompt_task.cancel()

    async def on_port_exit(self, port: PortClient, player: PlayerClient):
        self.game.update_port(port)
        self.game.update_player(player)
        self.game.player.port_id = None

        self.prompt = self.start_sector_prompt()
        self.prompt_task.cancel()

    async def on_invalid_action(self, error: str):
        self.term.error(error)
        self.prompt = self.start_sector_prompt()
        self.prompt_task.cancel()

    def quit(self):
        self.prompt = None
        self.prompt_task.cancel()

    def start_sector_prompt(self):
        actions = self.bus.wire_sending_methods(sector_prompt.Actions)()
        prompt = sector_prompt.Prompt(self.game, actions, term=self.term)
        prompt.print_sector()
        self.bus.append_event_listener(prompt)
        return prompt

    def start_port_prompt(self):
        actions = self.bus.wire_sending_methods(port_prompt.Actions)()
        prompt = port_prompt.Prompt(self.game, actions, self.term)
        events = port_prompt.Events(prompt)
        self.bus.append_event_listener(events)
        return prompt

    def start_no_prompt(self):
        prompt = InstantCmd(self.term)

        async def raise_quit(*_):
            raise PromptTransition(PromptType.QUIT)

        prompt.literal("q", default=False)(raise_quit)
        return prompt


class NoOpPrompt:
    async def cmdloop(self):
        fut = Future()
        await fut

    def __str__(self):
        return "No prompt"



import textwrap
from functools import partial
from typing import Awaitable, List, Set
from typing import Callable

from prompt_toolkit import Application
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.layout import D
from prompt_toolkit.layout import HSplit
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout import VSplit
from prompt_toolkit.widgets import Frame
from prompt_toolkit.widgets import Label
from prompt_toolkit.widgets import MenuContainer
from prompt_toolkit.widgets import MenuItem

from tspace.client.models import CommodityType, Sector
from tspace.client.session import Session
from tspace.client.stream import Terminal
from tspace.client.terminal_text_area import TerminalTextArea
from tspace.client.ui.dynamic_label import DynamicLabel
from tspace.client.ui.stat_frame import StatFrame, Stat


class TerminalScene:
    def __init__(self, app: Application, writer: Callable[[str], Awaitable[None]]):
        self.app = app
        self.writer = writer
        textfield = TerminalTextArea(focus_on_click=True, width=D(min=70))
        self.buffer = textfield.buffer
        self.buffer.on_change(lambda: app.invalidate())

        self.terminal = Terminal(self.buffer)
        self.session = Session(self.terminal)

        has_ship = lambda: bool(
            self.session
            and self.session.game
            and self.session.game.player
            and self.session.game.player.ship
        )

        root_container = HSplit(
            [
                VSplit(
                    [
                        HSplit(
                            [
                                StatFrame(
                                    condition=has_ship,
                                    stats=self.get_player_stats(),
                                    title="Player",
                                    width=D(max=25),
                                ),
                                StatFrame(
                                    condition=has_ship,
                                    stats=self.get_holds_stats(),
                                    title="Holds",
                                    width=D(max=25),
                                ),
                                StatFrame(
                                    condition=has_ship,
                                    stats=self.get_ship_stats(),
                                    title="Ship",
                                    width=D(max=25),
                                ),
                            ]
                        ),
                        textfield,
                        Frame(
                            title="Map",
                            body=DynamicLabel(partial(self.get_warps_label, has_ship)),
                            width=D(max=25),
                        ),
                    ],
                    height=D(),
                ),
                VSplit(
                    [
                        Frame(body=Label(text="right bottom frame\ncontent")),
                        Frame(body=Label(text="middle frame\ncontent")),
                        Frame(body=Label(text="left bottom frame\ncontent")),
                    ],
                    padding=1,
                ),
            ],
            style="bg:black",
        )

        def do_exit():
            self.session.quit()

        menu_container = MenuContainer(
            body=root_container,
            menu_items=[
                MenuItem(
                    "File",
                    children=[MenuItem("New"), MenuItem("Exit", handler=do_exit)],
                ),
                MenuItem(
                    "Edit",
                    children=[MenuItem("Undo"), MenuItem("Cut"), MenuItem("Copy")],
                ),
                MenuItem("View", children=[MenuItem("Status Bar")]),
                MenuItem("Info", children=[MenuItem("About")]),
            ],
        )
        self.layout = Layout(menu_container, focused_element=textfield)
        self.textfield = textfield

    def get_player_stats(self) -> List[Stat]:
        player = lambda: self.session.game.player
        return [
            Stat(title="Name", callable=lambda: player().name),
            Stat(
                title="Port",
                callable=lambda: "" if not player().port else player().port.class_name,
            ),
            Stat(title="Credits", callable=lambda: player().credits),
        ]

    def get_holds_stats(self) -> List[Stat]:
        ship = lambda: self.session.game.player.ship
        return [
            Stat(
                title="Total",
                callable=lambda: ship().holds_capacity - ship().holds_free,
            ),
            Stat(
                title="Fuel Ore", callable=lambda: ship().holds[CommodityType.fuel_ore]
            ),
            Stat(
                title="Organics", callable=lambda: ship().holds[CommodityType.organics]
            ),
            Stat(
                title="Equipment",
                callable=lambda: ship().holds[CommodityType.equipment],
            ),
            Stat(title="Empty", callable=lambda: ship().holds_free),
        ]

    def get_ship_stats(self) -> List[Stat]:
        ship = lambda: self.session.game.player.ship
        return [
            Stat(title="Name", callable=lambda: ship().name),
            Stat(title="Type", callable=lambda: ship().type),
            Stat(
                title="Weapons",
                multiline=True,
                callable=lambda: [w.name for w in ship().weapons],
            ),
            Stat(
                title="Countermeasures",
                multiline=True,
                callable=lambda: [w.name for w in ship().countermeasures],
            ),
        ]

    def get_warps_label(self, condition: Callable[[], bool]):
        frags = []
        if condition():
            sector = self.session.game.player.sector
            rows = self._calculate_max_rows(sector)

            if len(rows) > 40:
                max_depth = 2
            else:
                max_depth = 3

            # frags.append(("", f"len: {len(rows)}\n"))
            # frags.append(("", "\n".join(textwrap.wrap(f"rows: {','.join([str(r) for r in rows])}", 15))))
            visited = set()
            visited.add(sector.id)
            frags += self.print_sector(sector)
            frags.append(("", "\n"))
            for idx, warp in enumerate(sector.warps):
                frags += self._append_sector_warps(
                    warp, max_depth, 1, visited, last=idx == len(sector.warps) - 1
                )
        return FormattedText(frags)

    def _calculate_max_rows(self, sector):
        rows = []
        visited = set()
        visited.add(sector.id)
        for d1 in sector.warps:
            not_visited = d1 not in visited
            if self.session.game.sectors[d1].warps:
                visited.add(d1)
            rows.append(d1)
            if not_visited:
                for d2 in self.session.game.sectors[d1].warps:
                    not_visited = d2 not in visited
                    if self.session.game.sectors[d2].warps:
                        visited.add(d2)
                    rows.append(d2)
                    if not_visited:
                        for d3 in self.session.game.sectors[d2].warps:
                            rows.append(d3)
        return rows

    def _append_sector_warps(
        self,
        sector_id,
        max_depth: int,
        depth: int,
        visited: Set[int],
        last: bool = False,
    ) -> List:
        frags = []
        not_visited = not sector_id in visited

        w = self.session.game.sectors[sector_id]
        show_children = depth < max_depth and w.warps and not_visited
        if show_children:
            visited.add(sector_id)

        frags.append(("grey", "\u2502 " * (depth - 1)))
        if last and not show_children:
            frags.append(("grey", "\u2514 "))
        else:
            frags.append(("grey", "\u251C "))

        frags += self.print_sector(w)

        if show_children:
            frags.append(("", "\n"))
            # frags.append(("", " " * depth))
            # frags.append(("gray", "\\\n"))

            for idx, child_warp_id in enumerate(w.warps):
                last = idx == len(w.warps) - 1
                frags += self._append_sector_warps(
                    child_warp_id, max_depth, depth + 1, visited, last
                )
        elif w.warps:
            frags.append(("gray", "\u2026"))
            frags.append(("", "\n"))
        else:
            frags.append(("", "\n"))
        return frags

    def print_sector(self, sector: Sector):
        frags = []
        if sector.id in self.session.game.player.visited:
            frags.append(("cyan", str(sector.id)))
        else:
            frags.append(("red", f"({sector.id})"))
        if sector.ports:
            frags.append(("", " "))
            frags.append(("magenta", "("))
            frags += [(f.style, f.text) for f in sector.ports[0].class_name_colored]
            frags.append(("magenta", ")"))
        return frags

    async def start(self):
        return await self.session.start(self.writer)

    def end(self):
        pass



from __future__ import annotations

from typing import Callable
from typing import List
from typing import Optional

from prompt_toolkit.application import get_app
from prompt_toolkit.filters import has_completions
from prompt_toolkit.formatted_text import is_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.focus import focus_next
from prompt_toolkit.key_binding.bindings.focus import focus_previous
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import D
from prompt_toolkit.layout import DynamicContainer
from prompt_toolkit.layout import Float
from prompt_toolkit.layout import FloatContainer
from prompt_toolkit.layout import FormattedTextControl
from prompt_toolkit.layout import HSplit
from prompt_toolkit.layout import Window
from prompt_toolkit.layout import WindowAlign
from prompt_toolkit.mouse_events import MouseEventType
from prompt_toolkit.widgets import Box
from prompt_toolkit.widgets import Button
from prompt_toolkit.widgets import Frame
from prompt_toolkit.widgets import Shadow


class MenuDialog(object):
    """
    Forked from Dialog to add:
    - Custom background
    - Horizontal buttons for a menu

    :param body: Child container object.
    :param title: Text to be displayed in the heading of the dialog.
    :param buttons: A list of `Button` widgets, displayed at the bottom.
    """

    def __init__(
        self,
        body,
        title="",
        buttons: Optional[List[Button]] = None,
        modal=True,
        width=None,
        background=None,
        on_dismiss: Callable[[], None] = None,
    ):
        assert is_formatted_text(title)
        assert buttons is None or isinstance(buttons, list)

        self.body = body
        self.title = title

        max_menu_item_length = max(len(x.text) for x in buttons) + 10
        buttons = (
            []
            if not buttons
            else [
                MenuButton(btn.text, handler=btn.handler, width=max_menu_item_length)
                for btn in buttons
            ]
        )

        # When a button is selected, handle left/right key bindings.
        buttons_kb = KeyBindings()
        if len(buttons) > 1:
            buttons_kb.add("up")(focus_previous)
            buttons_kb.add("down")(focus_next)

        if buttons:
            frame_body = HSplit(
                [
                    # Add optional padding around the body.
                    Box(
                        body=DynamicContainer(lambda: self.body),
                        padding=D(preferred=1, max=1),
                        padding_bottom=1,
                    ),
                    # The buttons.
                    Box(
                        body=HSplit(buttons, padding=1, key_bindings=buttons_kb),
                        height=D(
                            min=len(buttons), max=20, preferred=len(buttons) * 2 + 1
                        ),
                        padding_bottom=1,
                    ),
                ]
            )
        else:
            frame_body = body

        # Key bindings for whole dialog.
        kb = KeyBindings()
        kb.add("tab", filter=~has_completions)(focus_next)
        kb.add("s-tab", filter=~has_completions)(focus_previous)

        if on_dismiss:
            kb.add("q")(on_dismiss)
            kb.add(Keys.Escape)(on_dismiss)

        frame = Shadow(
            body=Frame(
                title=lambda: self.title,
                body=frame_body,
                style="class:dialog.body",
                width=(None if background is None else width),
                key_bindings=kb,
                modal=modal,
            )
        )

        if background:
            self.container = FloatContainer(
                content=background,
                floats=[Float(transparent=False, content=Box(body=frame, width=width))],
            )
        else:
            self.container = frame

        # self.container = Starfield()

    def __pt_container__(self):
        return self.container


class MenuButton:
    """
    Clickable button.  Copied from Button but with prompt turned off

    :param text: The caption for the button.
    :param handler: `None` or callable. Called when the button is clicked.
    :param width: Width of the button.
    """

    def __init__(self, text, handler=None, width=20):
        assert isinstance(text, str)
        assert handler is None or callable(handler)
        assert isinstance(width, int)

        self.text = text
        self.handler = handler
        self.width = width
        self.control = FormattedTextControl(
            self._get_text_fragments,
            key_bindings=self._get_key_bindings(),
            show_cursor=False,
            focusable=True,
        )

        def get_style():
            if get_app().layout.has_focus(self):
                return "class:button.focused"
            else:
                return "class:button"

        self.window = Window(
            self.control,
            align=WindowAlign.CENTER,
            height=1,
            width=width,
            style=get_style,
            dont_extend_width=True,
            dont_extend_height=True,
        )

    def _get_text_fragments(self):
        text = ("{:^%s}" % (self.width - 10)).format(self.text)

        def handler(mouse_event):
            if mouse_event.event_type == MouseEventType.MOUSE_UP:
                self.handler()

        return [
            ("class:button.arrow", "--== ", handler),
            ("[SetCursorPosition]", ""),
            ("class:button.text", text, handler),
            ("class:button.arrow", " ==--", handler),
        ]

    def _get_key_bindings(self):
        " Key bindings for the Button. "
        kb = KeyBindings()

        @kb.add(" ")
        @kb.add("enter")
        def _(event):
            if self.handler is not None:
                self.handler()

        return kb

    def __pt_container__(self):
        return self.window



from typing import Awaitable
from typing import Callable
from typing import Dict

from tspace.server.config import GameConfig
from tspace.server.moves import GameConfigPublic
from tspace.server.moves import PlayerPublic
from tspace.server.moves import ShipMoves, ServerEvents
from tspace.server.galaxy import Galaxy
from tspace.server.util import CallMethodOnEventType


class Server:
    def __init__(self, config: GameConfig):
        self.config = config
        self.sessions: Dict[int, ServerEvents] = {}
        self.game = Galaxy(config)
        self.game.start()

    async def join(
        self, name, callback: Callable[[str], Awaitable[None]]
    ) -> Callable[[str], Awaitable[None]]:
        player = self.game.add_player(name)
        events = ServerEvents(callback)
        moves = ShipMoves(self, player, self.game, events)
        await events.on_game_enter(
            player=PlayerPublic(player),
            config=GameConfigPublic(self.config, len(self.game.sectors)),
        )
        self.sessions[player.id] = events
        await moves.broadcast_player_enter_sector(player)

        return CallMethodOnEventType(moves)



import inspect
import json
import types
from functools import wraps
from typing import Any

from tspace import json_types


class AutoIncrementId:
    def __init__(self):
        self.id = 1

    def incr(self):
        result = self.id
        self.id += 1
        return result


def methods_to_json(sink_property="target"):
    autoid = AutoIncrementId()

    def sender(func):
        @wraps(func)
        async def inner(self, **kwargs):
            target = getattr(self, sink_property)
            obj = {"type": func.__name__, "id": autoid.incr()}
            obj.update(kwargs)
            data = json_types.encodes(obj, exclude_none=True, set_as_list=True)
            resp = target(data)
            if not inspect.isawaitable(resp):
                breakpoint()
            await resp

        return inner

    def decorate(cls):
        for name, fn in {
            name: fn
            for name, fn in inspect.getmembers(cls)
            if isinstance(fn, types.FunctionType) and name.startswith("on_")
        }.items():
            setattr(cls, name, sender(fn))
        return cls

    return decorate


class CallMethodOnEventType:
    def __init__(self, *target: Any, context=None):
        self.targets = list(target)
        self.context = context

    def append(self, target: Any):
        if target is None:
            breakpoint()
        if target not in self.targets:
            self.targets.append(target)

    def remove(self, target: Any):
        if target in self.targets:
            self.targets.remove(target)

    async def __call__(self, data: str):
        event = json.loads(data)
        event_type = event["type"]
        for target in self.targets:
            try:
                func = getattr(target, event_type)
            except AttributeError:
                continue

            if func:
                kwargs = json_types.decode(event, func, context=self.context)
                kwargs["parent_id"] = event["id"]
                return await func(**kwargs)

        raise ValueError(f"No listeners found for {event_type} in {self.targets}")


class AutoIdDict(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.autoid = AutoIncrementId()

    def append(self, val):
        key = self.autoid.incr()
        self[key] = val
        val.id = key
        return val



from __future__ import annotations

import traceback
import typing
from typing import Callable

from tspace import json_types

from tspace.server.config import GameConfig
from tspace.server.models import CommodityType, Weapon, Countermeasure

from tspace.server.models import Planet
from tspace.server.models import Player
from tspace.server.models import Port
from tspace.server.models import Sector
from tspace.server.models import Ship
from tspace.server.models import TradingCommodity
from tspace.server.util import methods_to_json

if typing.TYPE_CHECKING:
    from tspace.server.server import Server
    from tspace.server.models import Galaxy


class GameConfigPublic:
    def __init__(self, config: GameConfig, sectors_count: int):
        self.id = config.id
        self.name = config.name
        self.diameter = config.diameter
        self.sectors_count = sectors_count


class PlayerPublic:
    def __init__(self, player: Player):
        self.id = player.id
        self.name = player.name
        self.ship = ShipPublic(player.ship, player.sector)
        self.credits = player.credits
        self.sector = SectorPublic(player.sector)
        self.port = None if not player.port else PortPublic(player.port)


class TraderPublic:
    def __init__(self, player: Player):
        self.id = player.id
        self.name = player.name


class TraderShipPublic:
    def __init__(self, ship: Ship):
        self.id = ship.id
        self.name = ship.name
        self.type = ship.ship_type
        self.trader = TraderPublic(ship.player)


class ShipPublic:
    def __init__(self, ship: Ship, sector: Sector):
        self.id = ship.id
        self.name = ship.name
        # self.type = ship.ship_type
        self.holds_capacity = ship.holds_capacity
        self.holds = {t.name: val for t, val in ship.holds.items()}
        self.sector = SectorPublic(sector)
        self.weapons = [WeaponPublic(w) for w in ship.weapons]
        self.countermeasures = [CountermeasurePublic(w) for w in ship.countermeasures]
        self.type = ship.ship_type.name


class WeaponPublic:
    def __init__(self, weapon: Weapon):
        self.id = weapon.id
        self.name = weapon.name
        self.damage_type = weapon.damage_type.name
        self.damage_min = weapon.damage_min
        self.damage_max = weapon.damage_max
        self.accuracy_bonus = weapon.accuracy_bonus


class CountermeasurePublic:
    def __init__(self, countermeasure: Countermeasure):
        self.id = countermeasure.id
        self.name = countermeasure.name
        self.strengths = [s.name for s in countermeasure.strengths]
        self.strength_bonus = countermeasure.strength_bonus
        self.weaknesses = [s.name for s in countermeasure.weaknesses]
        self.weakness_penalty = countermeasure.weakness_penalty


class TradingCommodityPublic:
    def __init__(self, commodity: TradingCommodity):
        self.amount = commodity.amount
        self.capacity = commodity.capacity
        self.buying = commodity.buying
        self.type = commodity.type.name
        self.price = commodity.price


class PortPublic:
    def __init__(self, port: Port):
        self.id = port.id
        self.name = port.name
        self.sector_id = port.sector_id
        self.commodities = [TradingCommodityPublic(c) for c in port.commodities]


class SectorPublic:
    def __init__(self, sector: Sector):
        self.id = sector.id
        self.coords = sector.coords
        self.warps = sector.warps
        self.ports = [PortPublic(port) for port in sector.ports]

        for port in self.ports:
            for c in port.commodities:
                c.amount = None
                c.capacity = None
                c.price = None
        self.ships = [TraderShipPublic(ship) for ship in sector.ships]
        self.planets = [PlanetPublic(planet) for planet in sector.planets]


class PlanetPublic:
    def __init__(self, planet: Planet):
        self.id = planet.id
        self.name = planet.name
        self.owner = TraderPublic(planet.owner) if planet.owner else None
        self.planet_type = planet.planet_type

        self.fuel_ore = 0
        self.organics = 0
        self.equipment = 0
        self.fighters = 0


@methods_to_json()
class ServerEvents:
    def __init__(self, target: Callable[[str], None]):
        self.target = target

    async def on_game_enter(self, player: PlayerPublic, config: GameConfigPublic):
        pass

    async def on_new_sector(self, sector: SectorPublic):
        pass

    async def on_ship_enter_sector(self, sector: SectorPublic, ship: TraderShipPublic):
        pass

    async def on_ship_exit_sector(self, sector: SectorPublic, ship: TraderShipPublic):
        pass

    async def on_invalid_action(self, error: str):
        pass

    async def on_port_enter(self, port: PortPublic, player: PlayerPublic):
        pass

    async def on_port_exit(self, port: PortPublic, player: PlayerPublic):
        pass

    async def on_port_buy(self, id: int, player: PlayerPublic):
        pass

    async def on_port_sell(self, id: int, player: PlayerPublic):
        pass

    async def reply(self, parent_id, *args):
        await self.target(
            json_types.encodes({"type": "reply", "parent_id": parent_id, "args": args})
        )


class ShipMoves:
    def __init__(
        self, server: Server, player: Player, galaxy: Galaxy, events: ServerEvents
    ):
        super().__init__()
        self.galaxy = galaxy
        self.player = player
        self.events = events
        self.server = server

    async def move_trader(self, parent_id: int, sector_id: int, **kwargs):
        if sector_id not in self.galaxy.sectors:
            await self.events.on_invalid_action(error="Not a valid sector number")
            return

        target = self.galaxy.sectors[sector_id]
        ship = self.galaxy.ships[self.player.ship_id]
        ship_sector = self.galaxy.sectors[ship.sector_id]

        if ship.player_id != self.player.id:
            await self.events.on_invalid_action(error="Ship not occupied by player")
            return

        if not ship_sector.can_warp(target.id):
            await self.events.on_invalid_action(
                error="Target sector not adjacent to ship"
            )
            return

        try:
            ship_sector.exit_ship(ship)
            target.enter_ship(ship)
            ship.move_sector(target.id)
            self.player.visit_sector(target.id)
            target_public = SectorPublic(target)
            await self.events.reply(parent_id, target_public)

            ship_as_trader = TraderShipPublic(ship)
            for ship in (s for s in ship_sector.ships if s.player_id != self.player.id):
                if ship.player_id in self.server.sessions:
                    await self.server.sessions[ship.player_id].on_ship_exit_sector(
                        sector=SectorPublic(ship_sector), ship=ship_as_trader
                    )

            await self.broadcast_ship_enter_sector(ship_as_trader, target_public)
        except Exception:
            traceback.print_exc()

    async def broadcast_player_enter_sector(self, player: Player):
        await self.broadcast_ship_enter_sector(
            TraderShipPublic(player.ship), SectorPublic(player.sector)
        )

    async def broadcast_ship_enter_sector(
        self, ship_as_trader: TraderShipPublic, target: SectorPublic
    ):
        for ship in (s for s in target.ships if s.trader.id != self.player.id):
            if ship.trader.id in self.server.sessions:
                await self.server.sessions[ship.trader.id].on_ship_enter_sector(
                    sector=target, ship=ship_as_trader
                )

    async def enter_port(self, port_id: int, **kwargs):
        port: Port = self.galaxy.ports[port_id]
        self.player.port = port
        await self.server.sessions[self.player.id].on_port_enter(
            port=PortPublic(port), player=PlayerPublic(self.player)
        )

    async def exit_port(self, port_id: int, **kwargs):
        port: Port = self.galaxy.ports[port_id]
        self.player.port = None
        await self.server.sessions[self.player.id].on_port_exit(
            port=PortPublic(port), player=PlayerPublic(self.player)
        )

    async def sell_to_port(
        self, parent_id, port_id: int, commodity: CommodityType, amount: int, **kwargs
    ):
        ship = self.galaxy.ships[self.player.ship_id]
        port = self.galaxy.ports[port_id]

        try:
            amount = int(amount)
        except ValueError:
            await self.events.on_invalid_action(error="Not a valid number")
            return

        trading = port.commodity(commodity)
        if not trading.buying:
            await self.events.on_invalid_action(
                error="This port is not buying that commodity"
            )
            return

        if trading.amount < amount:
            await self.events.on_invalid_action(error="Not that many available")
            return

        if ship.holds.get(commodity, 0) < amount:
            await self.events.on_invalid_action(error="Not enough goods in your holds")
            return

        cost = int(trading.price * amount)
        self.player.credits += cost
        trading.amount -= amount
        ship.remove_from_holds(commodity, amount)

        await self.events.reply(parent_id, PlayerPublic(self.player), PortPublic(port))

    async def buy_from_port(
        self,
        parent_id: int,
        port_id: int,
        commodity: CommodityType,
        amount: int,
        **kwargs,
    ):
        ship = self.galaxy.ships[self.player.ship_id]
        port = self.galaxy.ports[port_id]

        try:
            amount = int(amount)
        except ValueError:
            await self.events.on_invalid_action(error="Not a valid number")
            return

        trading = port.commodity(commodity)
        if trading.buying:
            await self.events.on_invalid_action(
                error="This port is not selling that commodity"
            )
            return

        if trading.amount < amount:
            await self.events.on_invalid_action(error="Not that many available")
            return

        cost = int(trading.price * amount)
        if cost > self.player.credits:
            await self.events.on_invalid_action(error="Not enough credits available")
            return

        if ship.holds_free < amount:
            await self.events.on_invalid_action(error="Not enough holds available")
            return

        self.player.credits -= cost
        trading.amount -= amount
        ship.add_to_holds(commodity, amount)

        await self.events.reply(parent_id, PlayerPublic(self.player), PortPublic(port))



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, List

from tspace.server.util import AutoIncrementId

from tspace.server.models import Sector

if TYPE_CHECKING:
    from tspace.server.models import Galaxy

autoid = AutoIncrementId()


def create(game: Galaxy, id: int, coords: Tuple[int, int], warps: List[int]) -> Sector:
    sector = Sector(game, id, coords, warps)
    game.sectors[sector.id] = sector
    return sector



import asyncio
from random import randrange

from prompt_toolkit.application import get_app
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import UIContent
from prompt_toolkit.layout import UIControl
from prompt_toolkit.layout import Window


class Starfield(UIControl):
    """
    Based on http://codentronix.com/2011/05/28/3d-starfield-made-using-python-and-pygame/
    """

    def __init__(self, num_stars: int = 256, max_depth: int = 12):
        self.num_stars = num_stars
        self.max_depth = max_depth

        self.key_bindings = KeyBindings()
        self.show_cursor = False

        self.window = Window(
            dont_extend_height=False,
            dont_extend_width=False,
            content=self,
            style="bg:black",
        )
        self.init_stars()

        async def tick():
            while True:
                get_app().invalidate()
                await asyncio.sleep(0.05)

        asyncio.create_task(tick())

    def init_stars(self):
        """ Create the starfield """
        self.stars = []
        for i in range(self.num_stars):
            # A star is represented as a list with this format: [X,Y,Z]
            star = [randrange(-5, 5), randrange(-5, 5), randrange(1, self.max_depth)]
            self.stars.append(star)

    def move_stars(self, width, height):
        """ Move and draw the stars """
        origin_x = width / 2
        origin_y = height / 2

        screen = [None] * height
        for y in range(height):
            row = [("", " ")] * width
            screen[y] = row

        for star in self.stars:
            # The Z component is decreased on each frame.
            star[2] -= 0.1

            # If the star has past the screen (I mean Z<=0) then we
            # reposition it far away from the screen (Z=max_depth)
            # with random X and Y coordinates.
            if star[2] <= 0:
                star[0] = randrange(-5, 5)
                star[1] = randrange(-5, 5)
                star[2] = self.max_depth

            # Convert the 3D coordinates to 2D using perspective projection.
            k = 128.0 / star[2]
            x = int(star[0] * k + origin_x)
            y = int(star[1] * k + origin_y)

            if 0 <= x < width and 0 <= y < height:
                size = (1 - float(star[2]) / self.max_depth) * 5
                shade = (1 - float(star[2]) / self.max_depth) * 255
                # print(f"star {id}: {x},{y}")

                hex_part = hex(round(shade))[2:]
                if len(hex_part) == 1:
                    hex_part = f"0{hex_part}"

                hex_shade = hex_part * 3

                dot = "."
                if size > 3.5:
                    parts = []
                    for _ in range(3):
                        val = hex(randrange(0, 255))[2:]
                        if len(val) == 1:
                            parts.append(f"0{val}")
                        else:
                            parts.append(val)
                    hex_shade = "".join(parts)
                    # print(f"shade: {hex_shade}")
                    dot = "\u26AB"
                elif size > 3:
                    dot = "o"
                elif size > 2:
                    dot = "*"
                # print(f"size:{size}")

                screen[y][x] = (f"#{hex_shade}", dot)
                # self.screen.fill((shade, shade, shade), (x, y, size, size))

        return screen

    def create_content(self, width, height):
        screen = self.move_stars(width, height)

        def get_line(i):
            result = screen[i]
            return result

        return UIContent(get_line=get_line, line_count=height)  # Something very big.

    def is_focusable(self):
        return False

    def __pt_container__(self):
        return self.window



from __future__ import annotations

import random
from typing import TYPE_CHECKING, Dict

import networkx

from tspace.server.graph import gen_hex_center, remove_warps
from tspace.server.models import (
    Sector,
    Player,
    Ship,
    Planet,
    Port,
    Weapon,
    Countermeasure,
)
from tspace.server.builders import planets, players, ports, sectors, ships

if TYPE_CHECKING:
    from tspace.server.config import GameConfig


class Galaxy:
    def __init__(self, config: GameConfig):
        self.config = config
        self.id = config.id
        self.name = config.name

        self.sectors: Dict[int, Sector] = {}
        self.ports: Dict[int, Port] = {}
        self.sector_coords_to_id = {}
        self.players: Dict[int, Player] = {}
        self.ships: Dict[int:Ship] = {}
        self.planets: Dict[int, Planet] = {}
        self.weapons: Dict[int, Weapon] = {}
        self.countermeasures: Dict[int, Countermeasure] = {}
        self._graph = None

        self.rnd = random.Random(self.config.seed)

    def add_player(self, name):

        p = players.create(self, name)

        sec = self.sectors[self.config.player.initial_sector_id]
        ship = ships.create_initial(self, "Foo", p.id, sec.id)

        p.ship_id = ship.id
        p.visit_sector(sec.id)
        ship.move_sector(sec.id)
        sec.enter_ship(ship)
        return p

    def id_to_coords(self, sector_id):
        return self.sectors[sector_id].coords

    def coords_to_id(self, x, y):
        return self.sector_coords_to_id[(x, y)]

    def _bang_world(self):
        g = gen_hex_center(self.config.diameter)
        remove_warps(g, self.config.warp_density, self.rnd)
        self._graph = g

        self.sector_coords_to_id = networkx.get_node_attributes(g, "sector_id")
        self.sector_coords_to_id[(0, 0)] = 1

        for n in g.nodes_iter():
            warps = [self.coords_to_id(*target) for target in g.neighbors(n)]
            sector_id = self.coords_to_id(*n)
            sector = sectors.create(self, sector_id, n, warps)

            if self.rnd.randint(1, 100) >= self.config.port.density:
                port = ports.create(self, sector_id)
                sector.port_ids.append(port.id)

            if self.rnd.randint(1, 2) == 1:
                for x in range(int(self.rnd.gauss(4.5, 1.5))):
                    planet = planets.create(self, None)
                    sector.planet_ids.append(planet.id)

    def start(self):
        self._bang_world()



import aiohttp
from aiohttp import web

from tspace.server.config import GameConfig
from tspace.server.server import Server


class WebGame:
    def __init__(self):
        self.config = GameConfig(
            1, "Test Game", diameter=10, seed="test", debug_network=False
        )
        self.server = Server(self.config)

    async def handler(self, request: web.Request):
        print("websocket connected")
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        async def cb(text):
            print(f"OUT: {text}")
            await ws.send_str(text)

        player_name = request.query["name"]

        in_cb = await self.server.join(player_name, cb)

        print("server joined")
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.text:
                print("IN: %s" % msg.data)
                await in_cb(msg.data)
            elif msg.type == aiohttp.WSMsgType.error:
                print("ws connection closed with exception %s" % ws.exception())
            else:
                print("unexpected message type: %s" % msg.type)

        print("websocket connection closed")

        return ws



from __future__ import annotations

from typing import TYPE_CHECKING

from tspace.server.models import Player
from tspace.server.util import AutoIncrementId

if TYPE_CHECKING:
    from tspace.server.models import Galaxy

autoid = AutoIncrementId()


def create(game: Galaxy, name: str) -> Player:
    player = Player(
        game, autoid.incr(), name, credits=game.config.player.initial_credits
    )
    game.players[player.id] = player
    return player



from functools import partial
from typing import Sequence
from typing import Tuple

from prompt_toolkit.application import get_app
from prompt_toolkit.filters import Condition, is_true
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import Margin
from prompt_toolkit.layout import Window, UIControl, ScrollbarMargin, UIContent
from prompt_toolkit.layout.screen import Point
from prompt_toolkit.mouse_events import MouseEventType

from tspace.client.twbuffer import TwBuffer


class TerminalTextArea(UIControl):
    """
    A simple input field.

    This is a higher level abstraction on top of several other classes with
    sane defaults.

    This widget does have the most common options, but it does not intend to
    cover every single use case. For more configurations options, you can
    always build a text area manually, using a
    :class:`~prompt_toolkit.buffer.Buffer`,
    :class:`~prompt_toolkit.layout.BufferControl` and
    :class:`~prompt_toolkit.layout.Window`.

    Buffer attributes:

    :param text: The initial text.
    :param multiline: If True, allow multiline input.
    :param completer: :class:`~prompt_toolkit.completion.Completer` instance
        for auto completion.
    :param complete_while_typing: Boolean.
    :param accept_handler: Called when `Enter` is pressed (This should be a
        callable that takes a buffer as input).
    :param history: :class:`~prompt_toolkit.history.History` instance.
    :param auto_suggest: :class:`~prompt_toolkit.auto_suggest.AutoSuggest`
        instance for input suggestions.

    BufferControl attributes:

    :param password: When `True`, display using asterisks.
    :param focusable: When `True`, allow this widget to receive the focus.
    :param focus_on_click: When `True`, focus after mouse click.
    :param input_processors: `None` or a list of
        :class:`~prompt_toolkit.layout.Processor` objects.

    Window attributes:

    :param lexer: :class:`~prompt_toolkit.lexers.Lexer` instance for syntax
        highlighting.
    :param wrap_lines: When `True`, don't scroll horizontally, but wrap lines.
    :param width: Window width. (:class:`~prompt_toolkit.layout.Dimension` object.)
    :param height: Window height. (:class:`~prompt_toolkit.layout.Dimension` object.)
    :param scrollbar: When `True`, display a scroll bar.
    :param style: A style string.
    :param dont_extend_width: When `True`, don't take up more width then the
                              preferred width reported by the control.
    :param dont_extend_height: When `True`, don't take up more width then the
                               preferred height reported by the control.
    :param get_line_prefix: None or a callable that returns formatted text to
        be inserted before a line. It takes a line number (int) and a
        wrap_count and returns formatted text. This can be used for
        implementation of line continuations, things like Vim "breakindent" and
        so on.

    Other attributes:

    :param search_field: An optional `SearchToolbar` object.
    """

    def __init__(
        self,
        text="",
        multiline=True,
        password=False,
        lexer=None,
        auto_suggest=None,
        completer=None,
        complete_while_typing=True,
        accept_handler=None,
        history=None,
        focusable=True,
        focus_on_click=False,
        wrap_lines=True,
        read_only=False,
        width=None,
        height=None,
        dont_extend_height=False,
        dont_extend_width=False,
        line_numbers=False,
        get_line_prefix=None,
        style="",
        search_field=None,
        preview_search=True,
        prompt="",
        input_processors=None,
    ):

        # Writeable attributes.

        self.buffer = TwBuffer()

        class MyMargin(Margin):
            def get_width(self, get_ui_content):
                return 2

            def create_margin(self, window_render_info, width, height):
                return [("", "  ")]

        right_margins = [ScrollbarMargin(display_arrows=True)]

        self.wrap_lines = wrap_lines

        left_margins = [MyMargin()]

        self.key_bindings = KeyBindings()
        self.show_cursor = True

        self.window = Window(
            height=height,
            width=width,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width,
            content=self,
            style=style,
            wrap_lines=Condition(lambda: is_true(self.wrap_lines)),
            left_margins=left_margins,
            right_margins=right_margins,
            get_line_prefix=get_line_prefix,
        )

        def handle_input(event: KeyPressEvent):
            self.show_cursor = True
            end_y = self.buffer.line_count - 1
            end_x = self.buffer.get_line_length(end_y) - 1
            if self.buffer.cursor.x != end_x or self.buffer.cursor.y != end_y:
                self.buffer.cursor = Point(end_x, end_y)
            txt = event.data
            if isinstance(event, KeyPressEvent):
                self.buffer.on_key_press(txt)
                # print(f"input: {txt}")
                # self.append_text([('', txt)])

        def scrollup(jump: int, _):
            # self.show_cursor = False
            self.buffer.cursor = Point(0, max(0, self.buffer.cursor.y - jump))

        def scrolldown(jump: int, _):
            # self.show_cursor = False
            y = min(self.buffer.line_count - 1, self.buffer.cursor.y + jump)
            x = 0
            if y == self.buffer.line_count - 1:
                x = self.buffer.get_line_length(y) - 1
            self.buffer.cursor = Point(
                x, min(self.buffer.line_count - 1, self.buffer.cursor.y + jump)
            )

        self.key_bindings.add(Keys.PageDown)(partial(scrolldown, 40))
        self.key_bindings.add(Keys.PageUp)(partial(scrollup, 40))
        self.key_bindings.add(Keys.Up)(partial(scrollup, 1))
        self.key_bindings.add(Keys.Down)(partial(scrolldown, 1))
        self.key_bindings.add(Keys.Any)(handle_input)

    def append_text(self, *text: Sequence[Tuple[str, str]]):
        self.buffer.insert_after(*text)

    def mouse_handler(self, mouse_event):
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            # Focus happens on mouseup. (If we did this on mousedown, the
            # up event will be received at the point where this widget is
            # focused and be handled anyway.)
            get_app().layout.current_control = self

    def get_key_bindings(self):
        return self.key_bindings

    # @property
    # def accept_handler(self):
    #     """
    #     The accept handler. Called when the user accepts the input.
    #     """
    #     return self.buffer.accept_handler
    #
    # @accept_handler.setter
    # def accept_handler(self, value):
    #     self.buffer.accept_handler = value

    def __pt_container__(self):
        return self.window

    def is_focusable(self):
        return True

    def create_content(self, width, height):
        return UIContent(
            get_line=self.buffer.get_line,
            line_count=self.buffer.line_count,
            cursor_position=self.buffer.cursor_position,
            show_cursor=self.show_cursor,
        )



from __future__ import annotations

import enum
import typing
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from tspace.client.stream import Fragment

if typing.TYPE_CHECKING:
    from tspace.client.game import Game


@dataclass
class GameConfigClient:
    id: int
    name: str
    diameter: int
    sectors_count: int


class GameConfig:
    def __init__(self, client: GameConfigClient):
        self.id = client.id
        self.name = client.name
        self.diameter = client.diameter
        self.sectors_count = client.sectors_count


class DamageType(enum.Enum):
    ENERGY = "Energy"
    KINETIC = "Kinetic"
    EXPLOSIVE = "Explosive"


@dataclass
class WeaponClient:
    id: int
    name: str
    accuracy_bonus: int
    damage_min: int
    damage_max: int
    damage_type: str


class Weapon:
    def __init__(self, game: Game, weapon: WeaponClient):
        self._game = game
        self.id: int = weapon.id
        self.update(weapon)

    def update(self, weapon: WeaponClient):
        self.name = weapon.name
        self.accuracy_bonus = weapon.accuracy_bonus
        self.damage_min: int = weapon.damage_min
        self.damage_max = weapon.damage_max
        self.damage_type = getattr(DamageType, weapon.damage_type.upper())


@dataclass
class CountermeasureClient:
    id: int
    name: str
    strengths: List[str]
    strength_bonus: int
    weaknesses: List[str]
    weakness_penalty: int


class Countermeasure:
    def __init__(self, game: Game, countermeasure: CountermeasureClient):
        self._game = game
        self.id: int = countermeasure.id
        self.update(countermeasure)

    def update(self, countermeasure: CountermeasureClient):
        self.name = countermeasure.name
        self.strengths = [getattr(DamageType, dt) for dt in countermeasure.strengths]
        self.strength_bonus = countermeasure.strength_bonus
        self.weaknesses = [getattr(DamageType, dt) for dt in countermeasure.weaknesses]
        self.weakness_penalty = countermeasure.weakness_penalty


@dataclass
class PlanetClient:
    id: int
    name: str
    regions: List
    planet_type: str
    fuel_ore: int
    organics: int
    equipment: int
    fighters: int
    owner: TraderClient


class Planet:
    def __init__(self, game: Game, planet: PlanetClient):
        self._game = game
        self.id: int = planet.id
        self.name: str = ""
        self.owner_id: int = 0
        self.planet_type: str = ""

        self.fuel_ore: int = 0
        self.organics: int = 0
        self.equipment: int = 0
        self.fighters: int = 0
        self.update(planet)

    def update(self, planet: PlanetClient):
        self.name = planet.name
        self.owner_id = planet.owner.id if planet.owner else None
        self.planet_type = planet.planet_type

        self.fuel_ore = planet.fuel_ore
        self.organics = planet.organics
        self.equipment = planet.equipment
        self.fighters = planet.fighters

    @property
    def owner(self):
        return self._game.traders[self.owner_id] if self.owner_id else None


@dataclass
class TradingCommodityClient:
    type: str
    buying: bool
    amount: int
    capacity: int
    price: int


class TradingCommodity:
    def __init__(self, client: TradingCommodityClient):
        self.type = CommodityType[client.type]
        self.price: int = None
        self.capacity: int = None
        self.amount: int = None
        self.buying: bool = False
        self.update(client)

    def update(self, client: TradingCommodityClient):
        self.price = client.price if client.price is not None else self.price
        self.capacity = (
            client.capacity if client.capacity is not None else self.capacity
        )
        self.amount = client.amount if client.amount is not None else self.amount
        self.buying = client.buying


class CommodityType(enum.Enum):
    fuel_ore = "Fuel Ore"
    organics = "Organics"
    equipment = "Equipment"


@dataclass
class PortClient:
    id: int
    sector_id: int
    name: str
    commodities: List[TradingCommodityClient]


class Port:
    CLASSES = {
        "BBS": 1,
        "BSB": 2,
        "SBB": 3,
        "SSB": 4,
        "SBS": 5,
        "BSS": 6,
        "SSS": 7,
        "BBB": 8,
    }

    def __init__(self, game: Game, client: PortClient):
        self._game = game
        self.id = client.id
        self.sector_id = client.sector_id
        self.name: str = ""
        self.commodities: Dict[CommodityType, TradingCommodity] = {}
        self.update(client)

    def update(self, client: PortClient):
        self.name = client.name

        if client.commodities:
            for commodity in client.commodities:
                commodity_type = CommodityType[commodity.type]
                if commodity_type in self.commodities:
                    self.commodities[commodity_type].update(commodity)
                else:
                    self.commodities[commodity_type] = TradingCommodity(commodity)

    @property
    def sector(self):
        return self._game.sectors[self.sector_id]

    @property
    def class_name(self):
        c = {c_type: c_value.buying for c_type, c_value in self.commodities.items()}

        name = []
        for ctype in CommodityType:
            name.append("B" if c[ctype] else "S")

        return "".join(name)

    @property
    def class_name_colored(self) -> List[Fragment]:
        line = []
        for c in self.class_name:
            line.append(Fragment("cyan", c) if c == "B" else Fragment("green", c))
        return line

    @property
    def class_number(self):
        return self.CLASSES[self.class_name]


@dataclass
class TraderClient:
    id: int
    name: str


class Trader:
    def __init__(self, game: Game, client: TraderClient):
        self._game = game
        self.id = client.id
        self.name: str = ""

    def update(self, client: TraderClient):
        self.name = client.name


@dataclass
class TraderShipClient:
    id: int
    name: str
    trader: TraderClient


class TraderShip:
    def __init__(self, game: Game, client: TraderShipClient):
        self.id = client.id
        self._game = game
        self.name: str = ""
        self.trader_id: int = 0

    def update(self, client: TraderShipClient):
        self.name = client.name
        self.trader_id = client.trader.id

    @property
    def trader(self) -> Optional[Trader]:
        return self._game.traders.get(self.trader_id)


@dataclass
class SectorClient:
    id: int
    warps: List[int]
    ports: List[PortClient]
    ships: List[TraderShipClient]
    planets: List[PlanetClient]


class Sector:
    def __init__(self, game: Game, client: SectorClient):
        self.id = client.id
        self._game = game
        self.update(client)

    def update(self, client: SectorClient):
        self.warps = client.warps

        self.port_ids = [port.id for port in client.ports]
        self.trader_ship_ids = [ship.id for ship in client.ships]
        self.planet_ids = [p.id for p in client.planets]

    @property
    def ports(self) -> List[Port]:
        return [self._game.ports[port_id] for port_id in self.port_ids]

    @property
    def ships(self) -> List[TraderShip]:
        return [
            self._game.trader_ships.get(ship_id) for ship_id in self.trader_ship_ids
        ]

    @property
    def planets(self) -> List[Planet]:
        return [self._game.planets.get(id) for id in self.planet_ids]


@dataclass
class ShipClient:
    id: int
    name: str
    type: str
    holds_capacity: int
    holds: Dict[str, int]
    weapons: List[WeaponClient]
    countermeasures: List[CountermeasureClient]
    sector: SectorClient


class Ship:
    def __init__(self, game: Game, client: ShipClient):
        self._game = game
        self.id = client.id
        self.name: str = ""
        self.holds_capacity: int = 0
        self.sector_id: int = 0

        self.holds: Dict[CommodityType, int] = defaultdict(lambda: 0)
        self.update(client)

    def update(self, client: ShipClient):
        self.name = client.name
        self.type = client.type
        self.holds_capacity = client.holds_capacity

        self.holds.update({CommodityType[k]: v for k, v in client.holds.items()})
        self.sector_id = client.sector.id if client.sector else None
        self.weapons = [Weapon(self._game, weapon) for weapon in client.weapons]
        self.countermeasures = [
            Countermeasure(self._game, cm) for cm in client.countermeasures
        ]

    @property
    def sector(self) -> Optional[Sector]:
        return None if not self.sector_id else self._game.sectors[self.sector_id]

    @property
    def holds_free(self):
        return self.holds_capacity - sum(self.holds.values())


@dataclass
class PlayerClient:
    id: int
    name: int
    credits: int
    ship: ShipClient


class Player:
    def __init__(self, game: Game, client: PlayerClient):
        self._game = game
        self.id = client.id
        self.visited: Set[int] = set()
        self.credits: int = 0
        self.name: str = ""
        self.ship_id: int = 0
        self.port_id: int = 0
        self.update(client)

    def update(self, client: PlayerClient):
        self.credits = client.credits
        self.name = client.name
        self.ship_id = client.ship.id if client.ship else None

    @property
    def ship(self) -> Optional[Ship]:
        return None if not self.ship_id else self._game.ships.get(self.ship_id)

    @property
    def port(self) -> Optional[Port]:
        return None if not self.port_id else self._game.ports.get(self.port_id)

    @property
    def sector(self) -> Optional[Sector]:
        return None if not self.ship_id else self.ship.sector



import random
import math

# from PIL import Image
# from PIL.ImageDraw import Draw
import networkx as nx
from networkx.algorithms import shortest_path_length


def gen_hex_center(size):
    g = nx.Graph(directed=False)
    g.add_node((0, 0))

    radius = math.ceil(size / 2)
    for rad in range(radius):
        for q, r in g.nodes():
            g.add_edge((q, r), (q, r - 1))
            g.add_edge((q, r), (q - 1, r))
            g.add_edge((q, r), (q - 1, r + 1))
            g.add_edge((q, r), (q, r + 1))
            g.add_edge((q, r), (q + 1, r - 1))
            g.add_edge((q, r), (q + 1, r))

    for x in range(radius):
        g.add_edge((x * -1, radius), (x * -1 - 1, radius))
        g.add_edge((x, -1 * radius), (x + 1, -1 * radius))
        g.add_edge((radius, x * -1), (radius, x * -1 - 1))
        g.add_edge((radius * -1, x), (-1 * radius, x + 1))
        g.add_edge((x * -1, radius * -1 + x), (x * -1 - 1, radius * -1 + x + 1))
        g.add_edge((x, radius - x), (x + 1, radius - x - 1))

    id_next = 1
    hit_nodes = set()
    for s, t in nx.bfs_edges(g, (0, 0)):
        if t not in hit_nodes:
            hit_nodes.add(t)
            id_next += 1
            g.add_node(t, sector_id=id_next)

    return g.to_directed()


# def gen_2d_grid(size):
#     g = nx.grid_2d_graph(size, size)
#     x = y = n = 0
#     for n in g:
#         x, y = n
#     if x > 0 and y > 0:
#         g.add_edge(n, (x - 1, y - 1))
#     if x < size - 1 and y < size - 1:
#         g.add_edge(n, (x + 1, y + 1))
#
#     return g.to_directed()


def remove_warps(g, density, rnd: random.Random):
    target_edge_count = int(g.number_of_nodes() * density)
    edges = nx.edges(g)
    removed_edges = []
    while len(edges) > target_edge_count:
        e = rnd.choice(edges)
        if len(g.neighbors(e[0])) > 1 and len(g.neighbors(e[1])) > 1:
            # print("Removing edge: {}".format(e))
            g.remove_edge(*e)
            edges.remove(e)
            other = e[::-1]
            g.remove_edge(*other)
            edges.remove(other)
            removed_edges += [e, other]
            # print("Removing edge: {} : {}".format(e, other))

    while True:
        connected = True
        for n in g.nodes_iter():
            try:
                shortest_path_length(g, source=(0, 0), target=n)
            except nx.exception.NetworkXNoPath:
                connected = False
                for e in list(removed_edges):
                    src, target = e
                    if target == n:
                        g.add_edge(src, target)
                        g.add_edge(target, src)
                        try:
                            shortest_path_length(g, source=(0, 0), target=n)
                            removed_edges.remove((src, target))
                            removed_edges.remove((target, src))
                            break
                        except nx.exception.NetworkXNoPath:
                            g.remove_edge(src, target)
                            g.remove_edge(target, src)
        if connected:
            break

    warps = [0] * 7
    for n in g.nodes_iter():
        try:
            shortest_path_length(g, source=(0, 0), target=n)
            warps[len(nx.neighbors(g, n))] += 1
        except nx.exception.NetworkXNoPath:
            raise Exception("bad")

    nodes_len = g.number_of_nodes()
    # print("Warps:")
    # for ind, num in enumerate(warps):
    #     print(" {} - {}%".format(ind, int((num / nodes_len) * 100)))
    # print('Done')


# def draw_hex_grid(g):
#     class HexagonGenerator(object):
#         """Returns a hexagon generator for hexagons of the specified size."""
#
#         def __init__(self, edge_length):
#             self.edge_length = edge_length
#
#         @property
#         def col_width(self):
#             return self.edge_length * 3
#
#         @property
#         def row_height(self):
#             return math.sin(math.pi / 3) * self.edge_length
#
#         def __call__(self, r, c):
#             x = (c + 0.5 * (r % 2)) * self.col_width
#             y = r * self.row_height
#             # noinspection PyArgumentList
#             for angle in range(0, 360, 60):
#                 x += math.cos(math.radians(angle)) * self.edge_length
#                 y += math.sin(math.radians(angle)) * self.edge_length
#                 yield x + self.edge_length, y
#
#     image = Image.new('RGB', (1500, 1000), 'white')
#     draw = Draw(image)
#     hexagon_generator = HexagonGenerator(40)
#     for n in g.nodes_iter():
#         row, col = n
#         hexagon = hexagon_generator(row, col)
#         draw.polygon(list(hexagon), outline='black', fill='red')
#
#     image.show()
#
#
# def draw_square_grid(g):
#     size = 10
#
#     def node_to_pixel(node):
#         return node[0] * (size * 2) + int(size / 2), node[1] * (size * 2) + int(size / 2)
#
#     size = 20
#     uni_width = math.sqrt(g.number_of_nodes())
#     image = Image.new('RGB', (size * 2 * uni_width, size * 2 * uni_width), 'white')
#
#     draw = Draw(image)
#     for e in g.edges_iter():
#         src, target = e
#         draw.line((node_to_pixel(src), node_to_pixel(target)), fill='blue', width=1)
#     for n in g.nodes_iter():
#         row, col = n
#         draw.ellipse((row * (size * 2), col * (size * 2), row * (size * 2) + size, col * size * 2 + size),
#                      outline='black', fill='red')
#     image.show()
#
#
# def draw_hex_circles(g, uni_radius):
#
#     size = 50
#     centerx = int(size * uni_radius * 2)
#     centery = int(size * uni_radius * 2)
#
#     def node_to_pixel(node):
#         x, y = node
#         x_offset = 0 if y == 0 else math.fabs(y) * size * (math.fabs(y) / y)
#         return x * (size * 2) + int(size / 2) + centerx + x_offset, y * (size * 2) + int(size / 2) + centery
#
#     image = Image.new('RGB', (size * 2 * uni_radius * 2, size * 2 * uni_radius * 2), 'white')
#
#     draw = Draw(image)
#     for e in g.edges_iter():
#         src, target = e
#         draw.line((node_to_pixel(src), node_to_pixel(target)), fill='blue', width=1)
#     for n in g.nodes_iter():
#         col, row = n
#         col_offset = 0 if row == 0 else math.fabs(row) * size * (math.fabs(row) / row)
#         draw.ellipse((col * (size * 2) + centerx + col_offset, row * (size * 2) + centery,
#                       col * (size * 2) + size + centerx + col_offset, row * size * 2 + size + centery),
#                      outline='black', fill='red')
#
#     for n in g.nodes_iter():
#         draw.text(node_to_pixel(n), str(n), fill='black')
#     image.show()
#
#
# if __name__ == "__main__":
#     uni_size = 35
#     density = 3.5
#
#     rnd = random.Random(2345)
#     g = gen_hex_center(10)
#     remove_warps(g, 3.5, rnd)
#     draw_hex_circles(g, uni_size)
#
#     #nx.draw_networkx(g, node_size=40)
#     #pyplot.show()
#     # draw_networkx(g)



from typing import Callable
from typing import Union

from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.formatted_text import fragment_list_to_text
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.layout import D
from prompt_toolkit.layout import FormattedTextControl
from prompt_toolkit.layout import Window
from prompt_toolkit.utils import get_cwidth


class DynamicLabel(object):
    """
    Widget that displays the given text. It is not editable or focusable.

    :param text: The text to be displayed. (This can be multiline. This can be
        formatted text as well.)
    :param style: A style string.
    :param width: When given, use this width, rather than calculating it from
        the text size.
    """

    def __init__(
        self,
        text_func: Callable[[], Union[str, FormattedText]],
        style="",
        width=None,
        dont_extend_height=True,
        dont_extend_width=False,
    ):
        self.text_func = text_func

        def get_width():
            if width is None:
                text_fragments = to_formatted_text(self.text_func())
                text = fragment_list_to_text(text_fragments)
                if text:
                    longest_line = max(get_cwidth(line) for line in text.splitlines())
                else:
                    return D(preferred=0)
                return D(preferred=longest_line)
            else:
                return width

        self.formatted_text_control = FormattedTextControl(text=text_func)

        self.window = Window(
            content=self.formatted_text_control,
            width=get_width,
            style="class:label " + style,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width,
        )

    def __pt_container__(self):
        return self.window



from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import List, Dict, Iterable, Tuple
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from tspace.server.galaxy import Galaxy


class Planet:
    def __init__(
        self,
        game: Galaxy,
        id: int,
        planet_type: str,
        name: str,
        owner_id: Optional[int],
    ):
        self.game: Galaxy = game
        self.name = name
        self.id = id
        self.owner_id = owner_id
        self.planet_type = planet_type

        self.fuel_ore = 0
        self.organics = 0
        self.equipment = 0
        self.fighters = 0

    @property
    def owner(self):
        return self.game.players[self.owner_id] if self.owner_id else None


class CommodityType(enum.Enum):
    fuel_ore = ("Fuel Ore", 1, 1.5)
    organics = ("Organics", 2, 3)
    equipment = ("Equipment", 4, 6)

    # noinspection PyInitNewSignature
    def __init__(self, title: str, sell_cost: int, buy_cost: float):
        self.title = title
        self.sell_cost = sell_cost
        self.buy_offer = buy_cost


class TradingCommodity:
    def __init__(self, type: CommodityType, amount: int, buying: bool):
        self.type: CommodityType = type
        self.amount = amount
        self.buying = buying
        self.capacity = amount

    @property
    def price(self):
        if self.buying:
            return round(
                self.type.buy_offer
                + ((self.amount / self.capacity) * self.type.buy_offer) / 2,
                2,
            )
        else:
            return round(
                self.type.sell_cost
                - ((self.amount / self.capacity) * self.type.sell_cost) / 2,
                2,
            )


class PortClass(enum.Enum):
    BBS = (1, True, True, False)
    BSB = (2, True, False, True)
    SBB = (3, False, True, True)
    SSB = (4, False, False, True)
    SBS = (5, False, True, False)
    BSS = (6, True, False, False)
    SSS = (7, False, False, False)
    BBB = (8, True, True, True)

    # noinspection PyInitNewSignature
    def __init__(self, id, buying_fuel_ore, buying_organics, buying_equipment):
        self.id = id
        self.buying = {
            CommodityType.fuel_ore: buying_fuel_ore,
            CommodityType.organics: buying_organics,
            CommodityType.equipment: buying_equipment,
        }

    @classmethod
    def by_id(cls, id: int):
        return next(e for e in cls if e.id == id)


class Port:
    def __init__(
        self, id: int, sector_id: int, name: str, commodities: List[TradingCommodity]
    ):
        self.id = id
        self.commodities = commodities
        self.name = name
        self.sector_id = sector_id

    def commodity(self, type: CommodityType):
        return next(c for c in self.commodities if c.type == type)


class Sector:
    def __init__(
        self, game: Galaxy, id: int, coords: Tuple[int, int], warps: List[int]
    ):
        self.game: Galaxy = game
        self.id = int(id)
        self.warps = [int(x) for x in warps]
        self.planet_ids: List[int] = []
        self.ship_ids = []
        self.coords = coords
        self.port_ids = []

    def can_warp(self, sector_id):
        return sector_id in self.warps

    def exit_ship(self, ship):
        self.ship_ids.remove(ship.id)

    def enter_ship(self, ship):
        self.ship_ids.append(ship.id)

    @property
    def ships(self):
        return [self.game.ships[id] for id in self.ship_ids]

    @property
    def planets(self):
        return [self.game.planets[id] for id in self.planet_ids]

    @property
    def ports(self):
        return [self.game.ports[id] for id in self.port_ids]


class Player:
    def __init__(self, game, id: int, name: str, credits: int):
        self.name = name
        self.id: int = id
        self.galaxy = game
        self.credits: int = credits
        self.ship_id = None
        self.port_id = None
        self.sector_id = None

    @property
    def sector(self):
        return self.galaxy.sectors[self.sector_id]

    @sector.setter
    def sector(self, value: Sector):
        self.sector_id = value.id

    @property
    def ship(self):
        return self.galaxy.ships[self.ship_id]

    @property
    def port(self):
        return self.galaxy.ports[self.port_id] if self.port_id else None

    @port.setter
    def port(self, value: Port):
        self.port_id = value.sector_id if value else None

    def visit_sector(self, sector_id):
        self.sector_id = sector_id

    def teleport(self, new_ship_id):
        self.ship_id = new_ship_id


@dataclass
class ShipType:
    name: str
    cost: int
    holds_initial: int
    holds_max: int
    warp_cost: int
    weapons_max: int
    countermeasures_max: int
    has_shield_slot: bool
    has_scanner_slot: bool


class DamageType(enum.Enum):
    ENERGY = "Energy"
    KINETIC = "Kinetic"
    EXPLOSIVE = "Explosive"


@dataclass
class Weapon:
    id: int
    name: str
    accuracy_bonus: int
    damage_min: int
    damage_max: int
    damage_type: DamageType


@dataclass
class Countermeasure:
    id: int
    name: str
    strengths: Iterable[DamageType]
    strength_bonus: int
    weaknesses: Iterable[DamageType]
    weakness_penalty: int


class Ship:
    def __init__(
        self,
        game: Galaxy,
        id: int,
        ship_type: ShipType,
        name: str,
        player_id: int,
        sector_id: int,
    ):
        self.id = id
        self.type = ship_type
        self.name = name
        self.player_owner_id = player_id
        self.player_id = player_id
        self.holds_capacity = ship_type.holds_initial
        self.holds: Dict[CommodityType, int] = {}
        self.ship_type = ship_type
        self.sector_id = sector_id
        self.game = game
        self.weapon_ids: List[int] = []
        self.countermeasure_ids: List[int] = []

    def move_sector(self, sector_id):
        self.sector_id = sector_id

    def add_to_holds(self, commodity_type: CommodityType, amount: int):
        self.holds[commodity_type] = self.holds.get(commodity_type, 0) + amount

    @property
    def holds_free(self):
        return self.holds_capacity - sum(self.holds.values())

    @property
    def player(self):
        return self.game.players[self.player_id]

    @property
    def sector(self):
        return self.game.sectors[self.sector_id]

    @property
    def weapons(self) -> List[Weapon]:
        return [self.game.weapons[id] for id in self.weapon_ids]

    @property
    def countermeasures(self) -> List[Countermeasure]:
        return [self.game.countermeasures[id] for id in self.countermeasure_ids]

    def remove_from_holds(self, commodity_type: CommodityType, amount: int):
        self.holds[commodity_type] -= amount

    def add_weapon(self, weapon: Weapon):
        if len(self.weapon_ids) < self.ship_type.weapons_max:
            self.weapon_ids.append(weapon.id)
        else:
            raise ValueError("Cannot add any more weapons")

    def add_countermeasure(self, countermeasure: Countermeasure):
        if len(self.countermeasure_ids) < self.ship_type.countermeasures_max:
            self.countermeasure_ids.append(countermeasure.id)
        else:
            raise ValueError("Cannot add any more countermeasures")



from __future__ import annotations

import re
from asyncio import coroutine
from inspect import isawaitable
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .stream import Terminal


class Matcher:
    def __init__(
        self,
        matcher: Callable[[str], bool],
        default=False,
        max_length=0,
        validator=None,
    ):
        self.default = default
        if validator is None:
            validator = lambda _: True
        self.validator: Callable[[str], bool] = validator
        self.func = matcher
        self.max_length = max_length

    def __call__(self, text):
        return self.validator(text) and self.func(text)


class InvalidSelectionError(Exception):
    pass


class InstantCmd:
    def __init__(self, out: Terminal):
        self.matchers: Dict[Matcher, Callable[[str], Any]] = {}
        self.out = out

    def regex(
        self,
        pattern,
        default=False,
        max_length=0,
        validator: Optional[Callable[[str], bool]] = None,
    ):
        return self.matcher(
            matcher=lambda txt: re.match(pattern, txt) is not None,
            default=default,
            max_length=max_length,
            validator=validator,
        )

    def literal(
        self, char, default=False, validator: Optional[Callable[[str], bool]] = None
    ):
        return self.matcher(
            matcher=lambda txt: txt.upper() == char.upper(),
            default=default,
            max_length=len(char),
            validator=validator,
        )

    def matcher(
        self,
        matcher: Callable[[str], bool],
        default: bool,
        max_length: int,
        validator: Optional[Callable[[str], bool]] = None,
    ):
        def outer(func: Callable[[str], Any]):
            self.matchers[
                Matcher(
                    matcher=matcher,
                    default=default,
                    max_length=max_length,
                    validator=validator,
                )
            ] = func

            async def inner(*args, **kwargs):
                result = func(*args, **kwargs)
                if isawaitable(result):
                    result = await result

                return result

            return inner

        return outer

    async def cmdloop(self):
        buffer = []

        while True:
            is_end = False
            should_write = False
            char = await self.out.read_key()
            # print(f"len: {len(char)} ord: {ord(char)}:", end='')
            if len(char) > 1:
                continue
            elif ord(char) == 127 and buffer:
                buffer = buffer[:-1]
                self.out.backspace(1)
                continue
            elif char == "\n" or char == "\r":
                is_end = True
            else:
                should_write = True
                buffer.append(char)

            line = "".join(buffer)
            matches = {k: v for k, v in self.matchers.items() if k(line)}

            if len(matches) == 1:
                matcher, func = next(iter(matches.items()))
                if should_write:
                    self.out.write_line(("", char))
                if matcher.max_length == len(buffer) or is_end:
                    return await coroutine(func)(line)
            elif is_end:
                try:
                    func = next(v for k, v in self.matchers.items() if k.default)
                except StopIteration:
                    buffer.clear()
                    continue

                if should_write:
                    self.out.write_line(("", char))
                return await coroutine(func)(line)
            elif not len(matches):
                buffer.pop()
            elif should_write:
                self.out.write_line(("", char))


# m = InstantCmd()
#
# @m.literal('f', default=True)
# def do_f(value):
#     return f"f: {value}"
#
# @m.literal('p')
# def do_p(value):
#     return f"p: {value}"
#
# @m.regex(r"[0-9]+", max_length=4)
# def do_num(value):
#     return f"num: {value}"
#
# result = m.cmdloop()
# print(f"result: {result}")



from __future__ import annotations

from typing import TYPE_CHECKING

from tspace.server.builders import weapons, countermeasures
from tspace.server.util import AutoIncrementId

from tspace.server.models import ShipType, Ship, DamageType

if TYPE_CHECKING:
    from tspace.server.models import Galaxy

autoid = AutoIncrementId()


MERCHANT_CRUISER = ShipType(
    name="Merchant Cruiser",
    cost=41300,
    holds_initial=200,
    holds_max=75,
    warp_cost=2,
    has_scanner_slot=True,
    has_shield_slot=True,
    weapons_max=3,
    countermeasures_max=3,
)


def create() -> Ship:
    breakpoint()


def create_initial(game: Galaxy, name: str, player_id: int, sector_id: int):
    ship = Ship(
        game,
        autoid.incr(),
        MERCHANT_CRUISER,
        name,
        player_id=player_id,
        sector_id=sector_id,
    )
    ship.add_weapon(weapons.create(game, DamageType.EXPLOSIVE))
    ship.add_weapon(weapons.create(game, DamageType.KINETIC))
    ship.add_weapon(weapons.create(game, DamageType.ENERGY))

    ship.add_countermeasure(countermeasures.create(game, DamageType.EXPLOSIVE))
    ship.add_countermeasure(countermeasures.create(game, DamageType.KINETIC))
    ship.add_countermeasure(countermeasures.create(game, DamageType.ENERGY))
    game.ships[ship.id] = ship
    return ship



import runpy
from os.path import splitext, basename, join, abspath, dirname

from setuptools import setup, find_packages, findall, find_namespace_packages
from setuptools.glob import glob

here = abspath(dirname(__file__))


def get_version():
    file_globals = runpy.run_path("tspace/version.py")
    return file_globals['__version__']


def read(*parts):
    with open(join(here, *parts), 'r') as f:
        return f.read()


LONG_DESCRIPTION = read('README.md')
REQUIREMENTS = read('requirements.txt')

setup(
    name='terminal-space',
    version=get_version(),
    description='A text-based space trading game',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Don Brown',
    author_email='mrdon@twdata.org',
    url='https://bitbucket.org/mrdon/terminal-space',
    license='aplv2',
    install_requires=REQUIREMENTS,
    packages=find_namespace_packages(include=['tspace.*']),
    include_package_data=True,
    entry_points="""
        [console_scripts]
        tspace-client = tspace.client_app:main
        tspace-server = tspace.server_app:main
    """,
    python_requires='>=3.7',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



import asyncio
from asyncio import CancelledError
from asyncio import Queue

import aiohttp
from prompt_toolkit import Application
from prompt_toolkit.eventloop import use_asyncio_event_loop
from prompt_toolkit.layout.screen import Size
from prompt_toolkit.styles import Style

from tspace.server.config import GameConfig
from tspace.server.server import Server
from tspace.client.scene.game import TerminalScene
from tspace.client.scene.main_menu import TitleScene


class InvalidScreenSize(Exception):
    def __init__(self, expected: Size, actual: Size):
        self.expected = expected
        self.actual = actual

    def __str__(self):
        return f"Invalid screen size: expected {self.expected} but was {self.actual}"


class TwApplication(Application):
    def __init__(self):
        super().__init__(
            mouse_support=True,
            full_screen=True,
            style=Style([("dialog", "bg:#fff333")]),
        )

        self.title_scene = TitleScene(self)
        self.layout = self.title_scene.layout

        actual: Size = self.output.get_size()
        expected = Size(rows=40, columns=120)
        if actual.rows < expected.rows or actual.columns < expected.columns:
            raise InvalidScreenSize(expected, actual)

    async def start(self):
        use_asyncio_event_loop()
        ui_task = self.run_async().to_asyncio_future()

        while True:
            self.layout = self.title_scene.layout
            self.invalidate()
            action = await self.title_scene.start()
            if action == "start":
                await self.start_game()
            elif action == "join":
                await self.join("localhost", "8080")
            elif action == "quit":
                self.exit(False)
                await ui_task
                break
            else:
                breakpoint()

            # self.exit(False)
        # await ui_task

    async def join(self, host: str, port: str):

        out_queue = Queue()

        terminal_scene = TerminalScene(self, lambda text: out_queue.put(text))
        self.layout = terminal_scene.layout

        async with aiohttp.ClientSession() as aiosession:
            async with aiosession.ws_connect(
                f"ws://{host}:{port}/?name=Remote%20Jim"
            ) as ws:

                async def read_input():
                    async for msg in ws:
                        if msg.type == aiohttp.WSMsgType.TEXT:
                            await terminal_scene.session.bus(msg.data)
                        elif msg.type == aiohttp.WSMsgType.ERROR:
                            print("error")
                            break

                async def write_output():
                    while True:
                        try:
                            msg = await out_queue.get()
                        except InterruptedError:
                            break
                        await ws.send_str(msg)

                write_task = asyncio.create_task(write_output())
                asyncio.create_task(read_input())

                try:
                    await terminal_scene.start()
                except CancelledError:
                    pass  # print("cancelled'")

                write_task.cancel()

        terminal_scene.end()

    async def start_game(self):
        config = GameConfig(
            1, "Test Game", diameter=10, seed="test", debug_network=False
        )
        server = Server(config)

        server_to_app = Queue()

        in_cb = await server.join(
            "Jim", lambda text: asyncio.coroutine(server_to_app.put_nowait)(text)
        )

        terminal_scene = TerminalScene(self, in_cb)
        self.layout = terminal_scene.layout

        async def read_from_server():
            while True:
                msg = await server_to_app.get()
                await terminal_scene.session.bus(msg)

        server_out_task = asyncio.create_task(read_from_server())

        try:
            await terminal_scene.start()
        except CancelledError:
            pass  # print("cancelled'")

        server_out_task.cancel()

        terminal_scene.end()


if __name__ == "__main__":

    async def run():
        app = TwApplication()
        await app.start()

    asyncio.get_event_loop().run_until_complete(run())



import inspect
import json
import types
import typing
from asyncio import Future
from asyncio import coroutine
from dataclasses import dataclass
from functools import wraps
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Iterable
from typing import Sequence
from typing import Tuple

from tspace import json_types


def frag_join(
    sep: Tuple[str, str], seq: Iterable[Tuple[str, str]]
) -> Sequence[Tuple[str, str]]:

    first = True
    for frag in seq:
        if not first:
            yield sep
        else:
            first = False
        yield frag


class AutoIncrementId:
    def __init__(self):
        self.id = 1

    def incr(self):
        result = self.id
        self.id += 1
        return result


@dataclass
class WaitForResponse:
    future: Future
    type: typing.Type


class EventBus:
    autoid = AutoIncrementId()

    def __init__(
        self, *target: Any, context=None, sender: Callable[[str], Awaitable[None]]
    ):
        self.targets = list(target)
        self.context = context
        self.sender = sender
        self.wait_for_ids: typing.Dict[int, WaitForResponse] = {}

    def wire_sending_methods(self, cls):
        out_self = self

        def sender(func):
            @wraps(func)
            async def inner(*_, **kwargs):
                target = out_self.sender
                event_id = out_self.autoid.incr()
                globalns = {
                    "Dict": typing.Dict,
                    "List": typing.List,
                    "Type": typing.Type,
                    "Tuple": typing.Tuple,
                }

                hints = typing.get_type_hints(
                    func, localns=self.context, globalns=globalns
                )
                # breakpoint()
                if hints and "return" in hints:
                    return_type = hints["return"]
                    self.wait_for_ids[event_id] = WaitForResponse(
                        type=return_type, future=Future()
                    )
                else:
                    return_type = None

                obj = {"type": func.__name__, "id": event_id}
                obj.update(kwargs)
                data = json_types.encodes(obj, exclude_none=True, set_as_list=True)
                resp = target(data)
                if not inspect.isawaitable(resp):
                    breakpoint()
                await resp

                if return_type:
                    return await self.wait_for_ids[event_id].future

            return inner

        for name, fn in {
            name: fn
            for name, fn in inspect.getmembers(cls)
            if isinstance(fn, types.FunctionType) and not name.startswith("__")
        }.items():
            setattr(cls, name, sender(fn))
        return cls

    def append_event_listener(self, target: Any):
        if target is None:
            breakpoint()
        if target not in self.targets:
            self.targets.append(target)

    def remove_event_listener(self, target: Any):
        if target in self.targets:
            self.targets.remove(target)

    async def __call__(self, data: str):
        event = json.loads(data)
        event_type = event["type"]
        event_id = event.get("parent_id")

        if event_id:
            args = event["args"]
            waited = self.wait_for_ids[event_id]
            origin = (
                waited.type.__origin__ if hasattr(waited.type, "__origin__") else None
            )
            if origin and issubclass(origin, tuple):
                result = []
                for t in waited.type.__args__:
                    obj = json_types.decode(args.pop(0), t, context=self.context)
                    result.append(obj)
                obj = (*result,)
            else:
                obj = json_types.decode(args.pop(0), waited.type, context=self.context)
            waited.future.set_result(obj)

        else:
            for target in self.targets:
                try:
                    func = getattr(target, event_type)
                except AttributeError:
                    continue

                if func:
                    kwargs = json_types.decode(event, func, context=self.context)
                    return await coroutine(func)(**kwargs)

            raise ValueError(f"No listeners found for {event_type} in {self.targets}")
